import ssl

from slixmpp import ClientXMPP
import logging

from slixmpp.exceptions import IqError, IqTimeout


class RegistrationException(Exception):
    pass


class XMPPClient(ClientXMPP):
    def __init__(self, jid, password, verify_security, auto_register):
        ClientXMPP.__init__(self, jid, password)

        self.logger = logging.getLogger("spade.Agent")

        if not verify_security:
            self.ssl_context.check_hostname = False
            self.ssl_context.verify_mode = ssl.CERT_NONE

        self.add_event_handler("session_start", self.session_start)
        self.register_plugin("xep_0199")  # XMPP Ping

        if auto_register:
            self.add_event_handler("register", self.register)
            self.register_plugin("xep_0077")  # In-band-registration

        self.register_plugin("xep_0199")  # Ping / Keepalive connection
        self["xep_0199"].enable_keepalive(interval=55)

    def session_start(self, event):
        self.send_presence()
        self.get_roster()

    async def register(self, event):
        resp = self.Iq()
        resp["type"] = "set"
        resp["register"]["username"] = self.boundjid.user
        resp["register"]["password"] = self.password

        try:
            await resp.send()
        except IqError:
            """
                If the user is already registered, it will return an IQ error
                We can safely ignore it. The client will try the auth process
                right after the ibr process
            """
            pass
        except IqTimeout:
            raise RegistrationException("Timeout error during the register process.")



import factory
from unittest.mock import AsyncMock, Mock

from spade.agent import Agent
from spade.presence import PresenceShow, PresenceType


class MockedConnectedAgent(Agent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._async_connect = AsyncMock()
        self._async_register = AsyncMock()
        self.stream = Mock()


class MockedAgentFactory(factory.Factory):
    class Meta:
        model = MockedConnectedAgent

    jid = "fake@jid"
    password = "fake_password"


class MockedPresenceConnectedAgent(Agent):
    def __init__(
        self, available=None, show=PresenceShow.NONE, status=None, priority=0, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if status is None:
            status = {}
        self._async_connect = AsyncMock()

        self.available = available
        self.show = show
        self.status = status
        self.priority = priority

    def mock_presence(self):
        show = self.show if self.show is not None else PresenceShow.NONE
        self.presence.set_presence(PresenceType.AVAILABLE, show, self.status, self.priority)


class MockedPresenceAgentFactory(factory.Factory):
    class Meta:
        model = MockedPresenceConnectedAgent

    jid = "fake@jid.com"
    password = "fake_password"
    available = None
    show = None
    status = {}
    priority = 0



import asyncio

import pytest
from slixmpp import JID, Iq

from spade.container import Container
from spade.message import Message


pytest_plugins = 'pytest_asyncio'


@pytest.fixture
def jid():
    return JID("friend@localhost/home")


@pytest.fixture
def message():
    return Message(
        to="to@localhost",
        sender="sender@localhost",
        body="message body",
        thread="thread-id",
        metadata={"metadata1": "value1", "metadata2": "value2"},
    )


@pytest.fixture
def message2():
    return Message(
        to="to2@localhost",
        sender="sender2@localhost",
        body="message body",
        thread="thread-id",
    )

@pytest.fixture
def iq():
    iq = Iq()
    iq['type'] = 'result'
    iq['id'] = '123'
    iq["to"] = "friend@localhost/home"
    iq['from'] = 'localhost'
    #set namespace to roster
    iq['roster']['xmlns'] = 'jabber:iq:roster'
    iq['type'] = 'result'
    iq['roster']['items'] = {
        'friend@localhost': {
            'name': 'My Friend',
            'subscription': 'both',
            'groups': ['Friends']
        },
        'friend2@localhost': {
            'name': 'User Two',
            'subscription': 'to',
            'groups': ['Work']
        }
    }
    return iq

# @pytest.fixture(autouse=True, scope="function")
# async def run_around_tests(event_loop):
#     container = Container()
#     container.reset()
#     if not container.is_running:
#         container.__init__()
#     container.loop = event_loop
#     yield
#     # Cancel tasks and close loop after testing
#     tasks = asyncio.all_tasks(loop=container.loop)
#     for task in tasks:
#         task.cancel()
#         from contextlib import suppress
#         with suppress(asyncio.CancelledError):
#             await task
    

@pytest.fixture(scope="module", autouse=True)
def cleanup(request):
    pass


async def wait_for_behaviour_is_killed(behaviour, tries=500, sleep=0.01):
    counter = 0
    while not behaviour.is_killed() and counter < tries:
        await asyncio.sleep(sleep)
        counter += 1
    if not behaviour.is_killed():
        raise Exception("Behaviour not finished")



import asyncio
from unittest.mock import Mock

import requests
from aiohttp import web
from aiohttp_jinja2 import get_env
from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
from parsel import Selector
from slixmpp import JID, Presence
from testfixtures import LogCapture

from spade.agent import Agent
from spade.behaviour import OneShotBehaviour, CyclicBehaviour
from spade.message import Message
from spade.presence import PresenceType, PresenceShow
from .factories import MockedAgentFactory, MockedPresenceAgentFactory


async def test_web():
    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.web.start(port=10000)

    assert agent.web.app is not None

    counter = 0
    while counter < 4:
        if agent.web.server is not None:
            break
        counter += 1
        await asyncio.sleep(0.1)
    assert agent.web.server is not None
    await agent.stop()


async def test_default_template_path():
    agent = Agent("jid@server", "password")

    env = get_env(agent.web.app)
    loader = env.loader

    package_loader = loader.loaders[0]
    filesystem_loader = loader.loaders[1]

    assert type(loader) == ChoiceLoader
    assert len(loader.loaders) == 2
    assert type(package_loader) == PackageLoader
    assert type(filesystem_loader) == FileSystemLoader

    assert "internal_tpl_agent.html" in package_loader.list_templates()
    assert "internal_tpl_agent.html" not in filesystem_loader.list_templates()
    assert filesystem_loader.searchpath == ["."]

    await agent.stop()


async def test_add_template_path_init():
    agent = Agent("jid@server", "password")

    agent.web.start(templates_path="/tmp/spade")

    env = get_env(agent.web.app)
    loader = env.loader

    assert type(loader) == ChoiceLoader
    assert len(loader.loaders) == 3
    assert type(loader.loaders[0]) == FileSystemLoader
    assert type(loader.loaders[1]) == PackageLoader
    assert type(loader.loaders[2]) == FileSystemLoader

    filesystem_loader = loader.loaders[0]

    assert filesystem_loader.list_templates() == []
    assert filesystem_loader.searchpath == ["/tmp/spade"]

    await agent.stop()


async def test_add_template_path():
    agent = Agent("jid@server", "password")

    agent.web.start()
    agent.web.add_template_path(templates_path="/tmp/spade")

    env = get_env(agent.web.app)
    loader = env.loader

    assert type(loader) == ChoiceLoader
    assert len(loader.loaders) == 3
    assert type(loader.loaders[0]) == FileSystemLoader
    assert type(loader.loaders[1]) == PackageLoader
    assert type(loader.loaders[2]) == FileSystemLoader

    filesystem_loader = loader.loaders[0]

    assert filesystem_loader.list_templates() == []
    assert filesystem_loader.searchpath == ["/tmp/spade"]

    await agent.stop()


async def test_check_server(aiohttp_client):
    agent = MockedPresenceAgentFactory()
    await agent.start(auto_register=False)

    agent.web.setup_routes()

    client = await aiohttp_client(agent.web.app)
    response = await client.get("/spade")

    assert response.status == 200

    text = await response.text()

    sel = Selector(text=text)

    assert sel.css("title::text").get() == "fake agent"
    assert sel.css("img::attr(src)").get() == agent.avatar

    assert sel.css("ul.products-list > li").getall() == []

    await agent.stop()


async def test_request_home(aiohttp_client):
    agent = MockedAgentFactory(jid="jid@server", password="password")
    await agent.start(auto_register=False)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.get("/spade")
    response = await response.text()

    sel = Selector(text=response)

    assert sel.css("title::text").get() == "jid agent"
    assert sel.css("img::attr(src)").get() == agent.avatar

    assert sel.css("ul.products-list > li").getall() == []

    await agent.stop()


async def test_get_messages(aiohttp_client):
    agent = Agent("jid@server", "password")
    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    # add messages to trace
    for i in range(5):
        msg = Message(body=str(i), sender="{}@server".format(i), to="receiver@server")
        agent.traces.append(msg)

    response = await client.get("/spade/messages/")
    response = await response.text()

    sel = Selector(text=response)

    assert len(sel.css("ul.timeline > li").getall()) == 6  # num messages + end clock

    await agent.stop()


async def test_get_behaviour(aiohttp_client):
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            self.kill()

    agent = Agent("jid@server", "password")
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)
    agent.web.setup_routes()

    client = await aiohttp_client(agent.web.app)

    response = await client.get(
        "/spade/behaviour/OneShotBehaviour/EmptyOneShotBehaviour/"
    )
    response = await response.text()

    sel = Selector(text=response)

    assert (
        sel.css("section.content-header > h1::text").get().strip()
        == "OneShotBehaviour/EmptyOneShotBehaviour"
    )
    await agent.stop()


async def test_kill_behaviour(aiohttp_client):
    class EmptyCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            await asyncio.sleep(0.1)

    agent = Agent("jid@server", "password")
    behaviour = EmptyCyclicBehaviour()
    agent.add_behaviour(behaviour)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    await client.get("/spade/behaviour/CyclicBehaviour/EmptyCyclicBehaviour/kill/")

    assert behaviour.is_killed()

    await agent.stop()


async def test_get_agent(aiohttp_client):
    agent = MockedPresenceAgentFactory(jid="jid@server.com/work", password="password")
    await agent.start(auto_register=False)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    jid = JID("friend@server.com/work")
    # Create a presence stanza with the desired change
    stanza = Presence()
    stanza["from"] = jid
    stanza["show"] = PresenceShow.CHAT.value
    # Trigger the 'changed_status' event with the presence stanza
    agent.client.event("changed_status", stanza)

    response = await client.get(f"/spade/agent/{jid.bare}/")
    response = await response.text()

    sel = Selector(text=response)

    assert sel.css("section.content-header > h1::text").get().strip() == jid.bare

    await agent.stop()


async def test_unsubscribe_agent(aiohttp_client, jid):
    agent = MockedPresenceAgentFactory()
    await agent.start(auto_register=False)

    agent.client.send_presence = Mock()

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    agent.client.update_roster(jid=jid, name="friend", subscription="both")

    response = await client.get(f"/spade/agent/unsubscribe/{jid}/")

    assert str(response.url.relative()) == "/spade"

    assert agent.client.send_presence.mock_calls
    arg = agent.client.send_presence.call_args[1]

    assert arg['pto'] == jid.bare
    assert arg['ptype'] == PresenceType.UNSUBSCRIBE.value

    await agent.stop()


async def test_send_agent(aiohttp_client, jid, iq):
    agent = MockedPresenceAgentFactory()
    await agent.start(auto_register=False)

    agent.client.send = Mock()
    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    agent.presence.handle_roster_update(iq)

    msg = "Hello World"

    response = await client.post(f"/spade/agent/{jid}/send/", data={"message": msg})

    assert response.status == 200
    assert str(response.url.relative()) == f"/spade/agent/{jid}/"

    sent = agent.traces.all()[0]

    assert sent[1].sent
    assert sent[1].body == "Hello World"

    await agent.stop()


async def test_find_behaviour():
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            pass

    agent = MockedAgentFactory()
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)
    found_behaviour = agent.web.find_behaviour("OneShotBehaviour/EmptyOneShotBehaviour")

    assert found_behaviour == behaviour

    await agent.stop()


async def test_find_behaviour_fail():
    agent = Agent("jid@server", "password")
    found_behaviour = agent.web.find_behaviour("OneShotBehaviour/EmptyOneShotBehaviour")

    assert found_behaviour is None

    await agent.stop()


async def test_add_get(aiohttp_client):
    agent = MockedAgentFactory()
    async def controller(request):
        return {"number": 42}
    agent.web.add_get("/test", controller, "tests/hello.html")

    await agent.start(auto_register=False)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.get("/test")

    assert response.status == 200

    response = await response.text()

    sel = Selector(text=response)
    assert sel.css("h1::text").get().strip() == "42"

    await agent.stop()


async def test_add_get_raw(aiohttp_client):
    agent = Agent("jid@server", "password")
    agent.web.add_get(
        "/test",
        lambda request: web.Response(text="Hello Raw Get"),
        template=None,
        raw=True,
    )

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.get("/test")
    response = await response.text()

    assert response == "Hello Raw Get"

    await agent.stop()


async def test_add_post(aiohttp_client):
    agent = Agent("jid@server", "password")

    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return {"number": number}

    agent.web.add_post("/test", handle_post, "tests/hello.html")
    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    response = await response.text()

    sel = Selector(text=response)
    assert sel.css("h1::text").get() == "1024"

    await agent.stop()


async def test_add_post_raw(aiohttp_client):
    agent = Agent("jid@server", "password")

    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return web.Response(text="Hello Raw Post Number={}".format(number))

    agent.web.add_post("/test", handle_post, template=None, raw=True)
    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    response = await response.text()

    assert response == "Hello Raw Post Number=1024"

    await agent.stop()


async def test_stop(aiohttp_client):
    agent = Agent("jid@server", "password")
    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.get("/spade/stop")
    assert response.status == 200

    response = await response.text()

    sel = Selector(text=response)
    assert (
        sel.css("div.alert-warning > span::text").get().strip()
        == "Agent is stopping now."
    )

    with LogCapture() as log:
        try:
            await client.get("/spade/stop/now/", timeout=0.05)
        except requests.exceptions.ReadTimeout:
            pass
        except asyncio.exceptions.TimeoutError:
            pass

        log.check_present(
            ("spade.Web", "WARNING", "Stopping agent from web interface.")
        )

    counter = 5
    while agent.is_alive() and counter > 0:
        counter -= 0.5
        await asyncio.sleep(0.5)

    assert not agent.is_alive()


async def test_add_get_json(aiohttp_client):
    async def controller(request):
        return {"number": 42}

    agent = Agent("jid@server", "password")
    agent.web.add_get("/test", controller, None)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.get("/test")
    assert response.status == 200

    data = await response.json()
    assert data["number"] == 42

    await agent.stop()


async def test_add_post_json(aiohttp_client):
    async def handle_post(request):
        form = await request.post()
        number = form["number"]
        return {"number": int(number)}

    agent = Agent("jid@server", "password")
    agent.web.add_post("/test", handle_post, None)

    agent.web.setup_routes()
    client = await aiohttp_client(agent.web.app)

    response = await client.post("/test", data={"number": 1024})
    assert response.status == 200

    data = await response.json()
    assert data["number"] == 1024

    await agent.stop()


async def test_add_menu_entry(aiohttp_client):
    agent = MockedPresenceAgentFactory()
    await agent.start(auto_register=False)
    agent.web.setup_routes()

    agent.web.add_menu_entry("Test", "/test", "fa fa-test")

    client = await aiohttp_client(agent.web.app)
    response = await client.get("/spade")

    assert response.status == 200

    # get html from response
    response = await response.text()
    # look for menu entry
    sel = Selector(text=response)

    assert len(sel.css("ul.sidebar-menu > li").getall()) == 2

    # get second li element from sidebar-menu
    menu_entry = sel.css("ul.sidebar-menu > li").getall()[1]
    menu_entry = Selector(text=menu_entry)
    assert menu_entry.css("li > a > span::text").get().strip() == "Test"
    assert menu_entry.css("a > i::attr(class)").get().strip() == "fa fa-test"
    assert menu_entry.css("a::attr(href)").get().strip() == "/test"



import datetime
import logging
import socket
from typing import Optional, Coroutine, Type

import aiohttp_jinja2
import jinja2
import timeago
from aiohttp import web as aioweb
from aiohttp.web_runner import AppRunner
from slixmpp import JID

from .behaviour import CyclicBehaviour
from .message import Message
from .presence import PresenceShow, ContactNotFound, PresenceNotFound

logger = logging.getLogger("spade.Web")


def unused_port(hostname: str) -> None:
    """Return a port that is unused on the current host."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((hostname, 0))
        return s.getsockname()[1]


async def start_server_in_loop(runner: AppRunner, hostname: str, port: int, agent):
    """
    Listens to http requests and sends them to the webapp.

    Args:
        runner (AppRunner): AppRunner to process the http requests
        hostname (str): host name to listen from.
        port (int): port to listen from.
        agent (spade.agent.Agent): agent that owns the web app.
    """
    await runner.setup()
    agent.web.server = aioweb.TCPSite(runner, hostname, port)
    await agent.web.server.start()
    logger.info(f"Serving on http://{hostname}:{port}/")


class WebApp(object):
    """Module to handle agent's web interface"""

    def __init__(self, agent):
        self.agent = agent
        self.server = None
        self.hostname = None
        self.port = None
        self.runner = None
        self.app = aioweb.Application()
        internal_loader = jinja2.PackageLoader(
            "spade", package_path="templates", encoding="utf-8"
        )
        cwd_loader = jinja2.FileSystemLoader(".")
        self.loaders = [internal_loader, cwd_loader]
        self._set_loaders()
        self.menu_entries = {}

    def start(
        self,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        templates_path: Optional[str] = None,
    ):
        """
        Starts the web interface.

        Args:
          hostname (str, optional): host name to listen from. (Default value = None)
          port (int, optional): port to listen from. (Default value = None)
          templates_path (str, optional): path to look for templates. (Default value = None)

        """
        self.hostname = hostname if hostname else "localhost"
        if port:
            self.port = port
        elif not self.port:
            self.port = unused_port(self.hostname)
        if templates_path:
            self.loaders.insert(0, jinja2.FileSystemLoader(templates_path))
            self._set_loaders()
        self.setup_routes()
        self.runner = aioweb.AppRunner(self.app)
        return self.agent.submit(
            start_server_in_loop(self.runner, self.hostname, self.port, self.agent)
        )

    def add_template_path(self, templates_path):
        self.loaders.insert(0, jinja2.FileSystemLoader(templates_path))
        self._set_loaders()

    def is_started(self) -> bool:
        return self.runner is not None

    def _set_loaders(self) -> None:
        loader = jinja2.ChoiceLoader(self.loaders)
        aiohttp_jinja2.setup(
            self.app,
            loader=loader,
            extensions=["jinja2_time.TimeExtension"],
            context_processors=[self.agent_processor, aiohttp_jinja2.request_processor],
        )

    def setup_routes(self) -> None:
        self.app.router.add_get("/spade", self.index)
        self.app.router.add_get("/spade/", self.index)
        self.app.router.add_get("/spade/stop", self.stop_agent)
        self.app.router.add_get("/spade/stop/now/", self.stop_now)
        self.app.router.add_get("/spade/messages/", self.get_messages)
        self.app.router.add_get(
            "/spade/behaviour/{behaviour_type}/{behaviour_class}/", self.get_behaviour
        )
        self.app.router.add_get(
            "/spade/behaviour/{behaviour_type}/{behaviour_class}/kill/",
            self.kill_behaviour,
        )

        self.app.router.add_get(
            "/spade/agent/{agentjid:[^/]+(?:/[^/]+)?}/", self.get_agent
        )
        self.app.router.add_get(
            "/spade/agent/unsubscribe/{agentjid:[^/]+(?:/[^/]+)?}/",
            self.unsubscribe_agent,
        )
        self.app.router.add_post(
            "/spade/agent/{agentjid:[^/]+(?:/[^/]+)?}/send/", self.send_agent
        )

    def add_menu_entry(self, name: str, url: str, icon="fa fa-circle") -> None:
        """
        Adds a new entry to the menu.

        Args:
          name (str): name of the entry
          url (str): url to be redirected to
          icon (str): icon to be displayed (Default value = "fa fa-circle")

        """
        self.menu_entries[name] = (url, icon)

    def add_get(
        self,
        path: str,
        controller: Coroutine,
        template: str,
        raw: Optional[bool] = False,
    ) -> None:
        """
        Setup a route of type GET

        Args:
          path (str): URL to listen to
          controller (coroutine): the coroutine to handle the request
          template (str): the template to render the response or None if it is a JSON response
          raw (bool): indicates if post-processing (jinja, json, etc) is needed or not

        """
        if raw:
            fn = controller
        else:
            fn = self._prepare_controller(controller, template)
        self.app.router.add_get(path, fn)

    def add_post(
        self,
        path: str,
        controller: Coroutine,
        template: str,
        raw: Optional[bool] = False,
    ) -> None:
        """
        Setup a route of type POST

        Args:
          path (str): URL to listen to
          controller (coroutine): the coroutine to handle the request
          template (str): the template to render the response or None if it is a JSON response
          raw (bool): indicates if post-processing (jinja, json, etc) is needed or not

        """
        if raw:
            fn = controller
        else:
            fn = self._prepare_controller(controller, template)
        self.app.router.add_post(path, fn)

    def _prepare_controller(self, controller: Coroutine, template: str) -> None:
        """
        Wraps the controller wether to render a jinja template or to return a json response (if template is None)
        Args:
            controller (coroutine): the coroutine to be wrapped
            template (str): the name of the template or None

        Returns:
            coroutine: a wrapped coroutine of the controller
        """
        if template:
            fn = aiohttp_jinja2.template(template_name=template)(controller)
        else:
            fn = self._parse_json_response(controller)
        return fn

    @staticmethod
    def _parse_json_response(func):
        async def wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            return aioweb.json_response(data=result, content_type="application/json")

        return wrapper

    @staticmethod
    def timeago(date):
        return timeago.format(date, datetime.datetime.now())

    async def agent_processor(self, request):
        messages = [
            (self.timeago(m[0]), m[1]) for m in self.agent.traces.received(limit=5)
        ]
        return {
            "agent": self.agent,
            "messages": messages,
            "menu": self.menu_entries,
            "active": request.path,
        }

    # Default controllers for agent

    @aiohttp_jinja2.template("internal_tpl_index.html")
    async def index(self, request):
        contacts = [
            {
                "jid": jid,
                "avatar": self.agent.build_avatar_url(jid),
                "available": c.is_available(),
                "show": c.get_presence().show,
                "resources": c.resources,
                "resource0": (
                    list(c.resources.values())[0].show,
                    list(c.resources.values())[0].status,
                    list(c.resources.values())[0].priority,
                ),
            }
            for jid, c in self.agent.presence.get_contacts().items()
        ]
        return {"contacts": contacts}

    @aiohttp_jinja2.template("internal_tpl_index.html")
    async def stop_agent(self, request):
        return {"stopping": True}

    async def stop_now(self, request):
        logger.warning("Stopping agent from web interface.")
        await self.agent.stop()
        return aioweb.json_response({})  # pragma: no cover

    @aiohttp_jinja2.template("internal_tpl_messages.html")
    async def get_messages(self, request):
        messages = [(self.timeago(m[0]), m[1]) for m in self.agent.traces.received()]
        return {"messages": messages}

    @aiohttp_jinja2.template("internal_tpl_behaviour.html")
    async def get_behaviour(self, request):
        behaviour_str = f"{request.match_info['behaviour_type']}/{request.match_info['behaviour_class']}"
        behaviour = self.find_behaviour(behaviour_str)
        messages = [
            (self.timeago(m[0]), m[1])
            for m in self.agent.traces.filter(category=behaviour_str)
        ]
        return {"behaviour": behaviour, "bmessages": messages}

    async def kill_behaviour(self, request):
        behaviour_str = f"{request.match_info['behaviour_type']}/{request.match_info['behaviour_class']}"
        behaviour = self.find_behaviour(behaviour_str)
        behaviour.kill()
        raise aioweb.HTTPFound("/spade")

    @aiohttp_jinja2.template("internal_tpl_agent.html")
    async def get_agent(self, request):
        agent_jid = request.match_info["agentjid"]
        agent_messages = [
            (self.timeago(m[0]), m[1]) for m in self.agent.traces.filter(to=agent_jid)
        ]
        try:
            c = self.agent.presence.get_contact(JID(agent_jid))
            contact = {"show": c.get_presence().show.value}
        except ContactNotFound:
            # raise 404
            raise aioweb.HTTPNotFound()
        except PresenceNotFound:
            contact = {"show": PresenceShow.NONE.value}

        except Exception as e:
            logger.error(e)

        return {"amessages": agent_messages, "ajid": agent_jid, "contact": contact}

    async def unsubscribe_agent(self, request):
        try:
            agent_jid = request.match_info["agentjid"]
            self.agent.presence.unsubscribe(str(JID(agent_jid).bare))
        except Exception as e:
            logger.error(e)
        raise aioweb.HTTPFound("/spade")

    async def send_agent(self, request):
        agent_jid = request.match_info["agentjid"]
        form = await request.post()
        body = form["message"]
        logger.info("Sending message to {}: {}".format(agent_jid, body))
        msg = Message(to=agent_jid, sender=str(self.agent.jid), body=body)
        slixmpp_msg = msg.prepare()
        self.agent.client.send(slixmpp_msg)
        msg.sent = True
        self.agent.traces.append(msg)
        raise aioweb.HTTPFound("/spade/agent/{agentjid}/".format(agentjid=agent_jid))

    def find_behaviour(self, behaviour_str: str) -> Optional[Type[CyclicBehaviour]]:
        behav = None
        for behaviour in self.agent.behaviours:
            if str(behaviour) == behaviour_str:
                behav = behaviour
                break
        return behav



from tests.factories import MockedConnectedAgent


async def test_add_plugin():
    class PluginMixin:
        async def _hook_plugin_before_connection(self, *args, **kwargs):
            await super()._hook_plugin_before_connection(*args, **kwargs)
            self.before_hook = True

        async def _hook_plugin_after_connection(self, *args, **kwargs):
            await super()._hook_plugin_after_connection(*args, **kwargs)
            self.after_hook = True

    class AgentWithPlugin(PluginMixin, MockedConnectedAgent):
        pass

    agent = AgentWithPlugin("test@localhost", "secret")

    await agent.start()

    assert agent.before_hook
    assert agent.after_hook


async def test_plugin_override():
    class PluginMixin:
        async def _hook_plugin_after_connection(self, *args, **kwargs):
            await super()._hook_plugin_after_connection(*args, **kwargs)
            self.client = None

    class AgentWithPlugin(PluginMixin, MockedConnectedAgent):
        pass

    agent = AgentWithPlugin("test@localhost", "secret")

    await agent.start()

    assert agent.client is None


async def test_plugin_multiple_override():
    """
    The order of the mixins is important
    """

    class PluginMixin1:
        async def _hook_plugin_after_connection(self, *args, **kwargs):
            await super()._hook_plugin_after_connection(*args, **kwargs)
            self.client = 1

    class PluginMixin2:
        async def _hook_plugin_after_connection(self, *args, **kwargs):
            await super()._hook_plugin_after_connection(*args, **kwargs)
            self.client = 2

    class PluginMixin3:
        async def _hook_plugin_after_connection(self, *args, **kwargs):
            await super()._hook_plugin_after_connection(*args, **kwargs)
            self.client = 3

    class AgentWithPlugin(
        PluginMixin1, PluginMixin2, PluginMixin3, MockedConnectedAgent
    ):
        pass

    agent = AgentWithPlugin("test@localhost", "secret")

    await agent.start()

    assert agent.client == 1



from unittest.mock import Mock
import pytest
from slixmpp.stanza import Presence, Iq
from slixmpp import JID

from spade.presence import ContactNotFound, PresenceShow, PresenceType, Contact
from .factories import MockedPresenceAgentFactory


async def test_get_state_not_available():
    agent = MockedPresenceAgentFactory(available=False, show=PresenceShow.NONE)

    await agent.start(auto_register=False)

    assert agent.presence.is_available() is False
    assert agent.presence. get_status() is None
    assert agent.presence.get_show() is PresenceShow.NONE


async def test_get_state_available():
    agent = MockedPresenceAgentFactory(available=True)

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.is_available() is True


async def test_set_presence_available():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_presence(PresenceType.AVAILABLE)

    assert agent.presence.is_available()


async def test_set_available_with_show():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_presence(PresenceType.AVAILABLE, show=PresenceShow.CHAT)

    assert agent.presence.is_available()
    assert agent.presence.get_show() == PresenceShow.CHAT


async def test_set_unavailable():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_unavailable()

    assert agent.presence.is_available() is False


async def test_get_state_show():
    agent = MockedPresenceAgentFactory(available=True, show=PresenceShow.AWAY)

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_show() == PresenceShow.AWAY


async def test_get_status_empty():
    agent = MockedPresenceAgentFactory(status="")

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_status() == ""


async def test_get_status_string():
    agent = MockedPresenceAgentFactory(status="Working")

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_status() == "Working"


async def test_get_status_dict():
    agent = MockedPresenceAgentFactory(status="Working")

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_status() == "Working"


async def test_get_priority_default():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_priority() == 0


async def test_get_priority():
    agent = MockedPresenceAgentFactory(priority=10)

    await agent.start(auto_register=False)

    agent.mock_presence()

    assert agent.presence.get_priority() == 10


async def test_set_presence_unavailable():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_unavailable()

    assert agent.presence.is_available() is False


async def test_set_presence_status():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_presence(status="Lunch")

    assert agent.presence.get_status() == "Lunch"


async def test_set_presence_priority():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_presence(priority=5)

    assert agent.presence.get_priority() == 5


async def test_set_presence():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.set_presence(show=PresenceShow.AWAY, status="Lunch", priority=2)

    assert agent.presence.is_available()
    assert agent.presence.get_show() == PresenceShow.AWAY
    assert agent.presence.get_status() == "Lunch"
    assert agent.presence.get_priority() == 2


async def test_get_contacts_empty():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    assert agent.presence.get_contacts() == {}


async def test_get_contacts(jid: JID, iq: Iq):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)
    
    agent.presence.handle_roster_update(iq)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare
    assert bare_jid in contacts
    assert len(contacts) == 2
    assert type(contacts[bare_jid]) == Contact
    assert contacts[bare_jid].name == "My Friend"
    assert contacts[bare_jid].subscription == "both"
    assert contacts[bare_jid].groups == ["Friends"]
    assert contacts[bare_jid].ask == "none"
    assert hasattr(contacts[bare_jid], "resources")
    assert isinstance(contacts[bare_jid].resources, dict)


async def test_get_contacts_with_update(jid: JID, iq: Iq):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.handle_roster_update(iq)

    stanza = Presence()
    stanza["from"] = jid
    stanza.set_show(PresenceShow.CHAT.value)
    stanza["status"] = "Just Chatting"
    stanza.set_priority(2)

    agent.presence.handle_presence(stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare
    assert bare_jid in contacts
    assert type(contacts[bare_jid]) == Contact
    assert contacts[bare_jid].name == "My Friend"
    assert contacts[bare_jid].subscription == "both"
    assert contacts[bare_jid].groups == ["Friends"]
    assert contacts[bare_jid].ask == "none"
    assert hasattr(contacts[bare_jid], "resources")
    assert isinstance(contacts[bare_jid].resources, dict)
    assert contacts[bare_jid].resources[jid.resource].type == PresenceType.AVAILABLE
    assert contacts[bare_jid].resources[jid.resource].show == PresenceShow.CHAT
    assert contacts[bare_jid].resources[jid.resource].status == "Just Chatting"
    assert contacts[bare_jid].resources[jid.resource].priority == 2

async def test_get_contacts_with_update_unavailable(jid: JID, iq: Iq):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.handle_roster_update(iq)

    stanza = Presence()
    stanza["from"] = jid
    stanza.set_type(PresenceType.UNAVAILABLE.value)

    agent.presence.handle_presence(stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare
    assert bare_jid in contacts
    assert type(contacts[bare_jid]) == Contact
    assert contacts[bare_jid].name == "My Friend"
    assert contacts[bare_jid].subscription == "both"
    assert contacts[bare_jid].groups == ["Friends"]
    assert contacts[bare_jid].ask == "none"
    assert hasattr(contacts[bare_jid], "resources")
    assert isinstance(contacts[bare_jid].resources, dict)
    assert contacts[bare_jid].resources[jid.resource].type == PresenceType.UNAVAILABLE
    assert contacts[bare_jid].resources[jid.resource].show == PresenceShow.NONE
    assert contacts[bare_jid].resources[jid.resource].status is None
    assert contacts[bare_jid].resources[jid.resource].priority == 0


async def test_get_contact(jid: JID, iq: Iq):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.handle_roster_update(iq)
    contact = agent.presence.get_contact(jid)

    assert type(contact) == Contact
    assert contact.name == "My Friend"
    assert contact.subscription == "both"
    assert len(contact.groups) == 1
    assert contact.ask == "none"
    assert hasattr(contact, "resources")
    assert isinstance(contact.resources, dict)


async def test_get_invalid_jid_contact():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    with pytest.raises(ContactNotFound):
        agent.presence.get_contact(JID("invalid@contact"))


async def test_get_invalid_str_contact():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    with pytest.raises(ContactNotFound):
        agent.presence.get_contact("invalid@contact")


async def test_subscribe(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.client.send_presence = Mock()
    agent.presence.subscribe(jid)

    assert agent.client.send_presence.mock_calls
    arg = agent.client.send_presence.call_args[1]

    assert arg["pto"] == jid
    assert arg["ptype"] == "subscribe"


async def test_unsubscribe(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.client.send_presence = Mock()
    agent.presence.unsubscribe(jid)

    assert agent.client.send_presence.mock_calls
    arg = agent.client.send_presence.call_args[1]

    assert arg["pto"] == jid
    assert arg["ptype"] == "unsubscribe"


async def test_approve(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.client.send_presence = Mock()
    agent.presence.approve_subscription(jid)

    assert agent.client.send_presence.mock_calls
    arg = agent.client.send_presence.call_args[1]

    assert arg["pto"] == jid
    assert arg["ptype"] == "subscribed"


async def test_on_available(jid: JID):
    import logging
    log = logging.getLogger("xmlstream")
    log.setLevel(logging.DEBUG)
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.on_available = Mock()

    stanza = Presence()
    stanza["from"] = jid
    stanza["type"] = PresenceType.AVAILABLE

    agent.client.event("presence_available", stanza)

    assert agent.presence.on_available.mock_calls

    assert len(agent.presence.on_available.mock_calls) == 2
    jid_arg = agent.presence.on_available.call_args[0][0]
    presence_info = agent.presence.on_available.call_args[0][1]
    last_presence = agent.presence.on_available.call_args[0][2]

    assert jid_arg == jid
    assert presence_info.type == PresenceType.AVAILABLE
    assert last_presence is None


async def test_on_unavailable(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.on_unavailable = Mock()

    stanza = Presence()
    stanza["from"] = jid
    stanza["type"] = "unavailable"

    agent.client.event("presence_unavailable", stanza)

    assert agent.presence.on_unavailable.mock_calls

    jid_arg = agent.presence.on_unavailable.call_args[0][0]
    presence_info = agent.presence.on_unavailable.call_args[0][1]
    last_presence = agent.presence.on_unavailable.call_args[0][2]

    assert jid_arg == jid
    assert presence_info.type == PresenceType.UNAVAILABLE
    assert last_presence is None


async def test_on_changed_status(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    # Create a presence stanza with the desired change
    stanza = Presence()
    stanza["from"] = jid
    stanza["show"] = PresenceShow.CHAT.value

    # Trigger the 'changed_status' event with the presence stanza
    agent.client.event("changed_status", stanza)
    agent.presence.subscribe(jid.bare)

    # Verify that the contact in the roster has the updated presence
    contacts = agent.presence.get_contacts()
    bare_jid = jid.bare

    # Check that the contact is in the roster and has the correct presence
    assert bare_jid in contacts
    contact = contacts[bare_jid]

    # Verify that the contact has updated presence information
    assert contact.resources is not None
    assert len(contact.resources) > 0

    # Get the presence information for the resource and verify the 'show' state
    resource_presence = contact.resources.get(jid.resource)
    assert resource_presence is not None
    assert resource_presence.show == PresenceShow.CHAT


async def test_on_subscribe(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.on_subscribe = Mock()

    stanza = Presence()
    stanza["from"] = jid
    stanza["type"] = PresenceType.SUBSCRIBE.value

    agent.client.event("presence_subscribe", stanza)

    assert agent.presence.on_subscribe.mock_calls

    jid_arg = agent.presence.on_subscribe.call_args[0][0]

    assert jid_arg == jid.bare


async def test_on_subscribed(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.on_subscribed = Mock()

    stanza = Presence()
    stanza["from"] = jid
    stanza["type"] = PresenceType.SUBSCRIBED.value

    agent.client.event("presence_subscribed", stanza)

    assert agent.presence.on_subscribed.mock_calls

    jid_arg = agent.presence.on_subscribed.call_args[0][0]

    assert jid_arg == jid.bare


async def test_on_unsubscribe(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.on_unsubscribe = Mock()

    stanza = Presence()
    stanza["from"] = jid
    stanza["type"] = PresenceType.UNSUBSCRIBE.value

    agent.client.event("presence_unsubscribe", stanza)

    assert agent.presence.on_unsubscribe.mock_calls

    jid_arg = agent.presence.on_unsubscribe.call_args[0][0]

    assert jid_arg == jid.bare


async def test_ignore_self_presence():
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    jid_ = agent.jid

    stanza = Presence()
    stanza["from"] = jid_
    stanza["type"] = PresenceType.AVAILABLE.value
    stanza["show"] = PresenceShow.CHAT.value

    agent.client.event("presence_available", stanza)

    with pytest.raises(ContactNotFound):
        agent.presence.get_contact(jid_)

    await agent.stop()



import asyncio
import datetime
from unittest.mock import AsyncMock, Mock

import pytest

from slixmpp import JID

from spade.agent import Agent
from spade.behaviour import (
    OneShotBehaviour,
    CyclicBehaviour,
    PeriodicBehaviour,
    TimeoutBehaviour,
    FSMBehaviour,
    State,
    NotValidState,
    NotValidTransition,
    BehaviourNotFinishedException,
)
from spade.message import Message, SPADE_X_METADATA
from spade.template import Template
from .conftest import wait_for_behaviour_is_killed
from .factories import MockedAgentFactory

STATE_ONE = "STATE_ONE"
STATE_TWO = "STATE_TWO"
STATE_THREE = "STATE_THREE"


class StateOne(State):
    async def run(self):
        self.agent.state = STATE_ONE
        self.set_next_state(STATE_TWO)
        self.kill()
        await self.agent.sync1_behaviour.wait()


class StateTwo(State):
    async def run(self):
        self.agent.state = STATE_TWO
        self.set_next_state(STATE_THREE)
        self.kill()
        await self.agent.sync2_behaviour.wait()


class StateThree(State):
    async def run(self):
        self.agent.state = STATE_THREE
        self.kill()


@pytest.fixture
def fsm():
    fsm_ = FSMBehaviour()
    state_one = StateOne()
    state_two = StateTwo()
    state_three = StateThree()
    fsm_.add_state(STATE_ONE, state_one, initial=True)
    fsm_.add_state(STATE_TWO, state_two)
    fsm_.add_state(STATE_THREE, state_three)
    fsm_.add_transition(STATE_ONE, STATE_TWO)
    fsm_.add_transition(STATE_TWO, STATE_THREE)

    fsm_.state_one = state_one
    fsm_.state_two = state_two
    fsm_.state_three = state_three

    return fsm_


async def test_on_start_on_end():
    class TestOneShotBehaviour(OneShotBehaviour):
        async def on_start(self):
            self.agent.on_start_flag = True

        async def run(self):
            pass

        async def on_end(self):
            self.agent.on_end_flag = True
            self.kill()

    agent = MockedAgentFactory()
    agent.on_start_flag = False
    agent.on_end_flag = False
    behaviour = TestOneShotBehaviour()
    agent.add_behaviour(behaviour)

    assert agent.on_start_flag is False
    assert agent.on_end_flag is False

    await agent.start(auto_register=False)

    await behaviour.join()

    assert agent.on_start_flag is True
    assert agent.on_end_flag is True
    await agent.stop()


async def test_on_start_exception():
    class TestOneShotBehaviour(OneShotBehaviour):
        async def on_start(self):
            1 / 0
            self.agent.flag = True

        async def run(self):
            pass

    agent = MockedAgentFactory()
    agent.flag = False
    behaviour = TestOneShotBehaviour()
    agent.add_behaviour(behaviour)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert type(behaviour.exit_code) == ZeroDivisionError
    assert not agent.flag
    await agent.stop()


async def test_on_run_exception():
    class TestOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            1 / 0
            self.agent.flag = True

    agent = MockedAgentFactory()
    agent.flag = False
    behaviour = TestOneShotBehaviour()
    agent.add_behaviour(behaviour)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert type(behaviour.exit_code) == ZeroDivisionError
    assert not agent.flag
    await agent.stop()


async def test_on_end_exception():
    class TestOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            pass

        async def on_end(self):
            1 / 0
            self.agent.flag = True

    agent = MockedAgentFactory()
    agent.flag = False
    behaviour = TestOneShotBehaviour()
    agent.add_behaviour(behaviour)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert type(behaviour.exit_code) == ZeroDivisionError
    assert not agent.flag
    await agent.stop()


async def test_add_behaviour():
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            self.kill()

    agent = MockedAgentFactory()
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)

    assert agent.has_behaviour(behaviour)

    await agent.start(auto_register=False)

    assert behaviour.agent == agent
    assert behaviour.template is None
    assert behaviour.presence == agent.presence
    assert behaviour.web == agent.web
    assert behaviour.queue.empty()

    await behaviour.join()

    assert behaviour.is_done()
    await agent.stop()


def test_remove_behaviour():
    class EmptyBehaviour(CyclicBehaviour):
        async def run(self):
            pass

    agent = MockedAgentFactory()
    behaviour = EmptyBehaviour()
    agent.add_behaviour(behaviour)
    assert agent.has_behaviour(behaviour)

    agent.remove_behaviour(behaviour)

    assert not agent.has_behaviour(behaviour)


def test_remove_behaviour_not_added():
    class EmptyBehaviour(CyclicBehaviour):
        async def run(self):
            pass

    agent = MockedAgentFactory()
    behaviour = EmptyBehaviour()

    with pytest.raises(ValueError):
        agent.remove_behaviour(behaviour)


async def test_wait_for_agent_start():
    class EmptyOneShotBehaviour(OneShotBehaviour):
        async def on_start(self):
            self.agent.started = True
            self.kill()

        async def run(self):
            pass

    agent = MockedAgentFactory()
    behaviour = EmptyOneShotBehaviour()
    agent.add_behaviour(behaviour)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert agent.started

    await agent.stop()


def test_behaviour_match():
    class TestBehaviour(OneShotBehaviour):
        async def run(self):
            pass

    template = Template()
    template.sender = "sender1@host"
    template.to = "recv1@host"
    template.body = "Hello World"
    template.thread = "thread-id"
    template.metadata = {"performative": "query"}

    behaviour = TestBehaviour()
    behaviour.set_template(template)

    msg = Message()
    msg.sender = "sender1@host"
    msg.to = "recv1@host"
    msg.body = "Hello World"
    msg.thread = "thread-id"
    msg.set_metadata("performative", "query")

    assert behaviour.match(msg)


def test_behaviour_match_without_template():
    class TestBehaviour(OneShotBehaviour):
        async def run(self):
            pass

    behaviour = TestBehaviour()

    msg = Message()
    msg.sender = "sender1@host"
    msg.to = "recv1@host"
    msg.body = "Hello World"
    msg.thread = "thread-id"
    msg.set_metadata("performative", "query")

    assert behaviour.match(msg)


async def test_send_message(message):
    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            await self.send(message)
            self.kill()

    agent = MockedAgentFactory(jid="sender@localhost")
    await agent.start(auto_register=False)

    agent2 = MockedAgentFactory(jid="to@localhost")
    await agent2.start(auto_register=False)

    agent2.dispatch = Mock()

    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)

    await behaviour.join()

    assert agent2.dispatch.assert_called
    msg_arg = agent2.dispatch.call_args[0][0]
    assert msg_arg.body == "message body"
    assert msg_arg.to == JID("to@localhost")
    assert msg_arg.thread == "thread-id"

    await agent.stop()
    await agent2.stop()


async def test_send_message_to_external_agent():
    message = Message(
        to="to@external.xmpp.com",
        sender="sender@localhost",
        body="message body",
        thread="thread-id",
        metadata={"metadata1": "value1", "metadata2": "value2"},
    )

    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            await self.send(message)
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.client.send = Mock()
    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)

    await behaviour.join()

    assert agent.client.send.call_count == 1
    msg_arg = agent.client.send.call_args[0][0]
    assert msg_arg['body'] == "message body"
    assert msg_arg['to'] == JID("to@external.xmpp.com")
    thread_found = False
    for data in [pl for pl in msg_arg.get_payload() if pl.tag == '{jabber:x:data}x']:
        if data.find('{jabber:x:data}title').text == SPADE_X_METADATA:
            for field in data.findall('{jabber:x:data}field'):
                if field.attrib['var'] == "_thread_node":
                    assert field.find('{jabber:x:data}value').text == "thread-id"
                    thread_found = True
    assert thread_found

    await agent.stop()


async def test_send_message_without_sender():
    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            msg = Message()
            await self.send(msg)
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.client.send = Mock()
    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)

    await behaviour.join()

    msg_arg = agent.client.send.call_args[0][0]
    assert msg_arg['from'] == JID("fake@jid")

    await agent.stop()


async def test_receive():
    class RecvBehaviour(OneShotBehaviour):
        async def run(self):
            msg = Message(body="received body")
            await asyncio.wait_for(self.queue.put(msg), 5.0)
            self.agent.recv_msg = await self.receive()
            self.kill()

    agent = MockedAgentFactory()

    behaviour = RecvBehaviour()
    agent.add_behaviour(behaviour)
    assert behaviour.mailbox_size() == 0

    await agent.start(auto_register=False)

    assert agent.is_alive()

    await behaviour.join()

    assert agent.recv_msg.body == "received body"

    await agent.stop()


async def test_receive_with_timeout():
    class RecvBehaviour(OneShotBehaviour):
        async def run(self):
            self.agent.recv_msg = await self.receive(15.0)
            self.kill()

    agent = MockedAgentFactory()

    msg = Message(body="received body")
    template = Template(body="received body")
    behaviour = RecvBehaviour()
    agent.add_behaviour(behaviour, template)
    assert behaviour.mailbox_size() == 0

    await agent.start(auto_register=False)

    assert agent.is_alive()
    assert agent.has_behaviour(behaviour)
    agent._message_received(msg.prepare())

    await behaviour.join()

    assert agent.recv_msg.body == "received body"
    assert agent.recv_msg == msg

    await agent.stop()


async def test_receive_with_timeout_error():
    class RecvBehaviour(OneShotBehaviour):
        async def run(self):
            self.agent.recv_msg = await self.receive(0.01)
            self.kill()

    agent = MockedAgentFactory()

    template = Template(body="received body")
    behaviour = RecvBehaviour()
    agent.add_behaviour(behaviour, template)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert behaviour.mailbox_size() == 0
    assert agent.recv_msg is None
    await agent.stop()


async def test_receive_with_empty_queue():
    class RecvBehaviour(OneShotBehaviour):
        async def run(self):
            self.agent.recv_msg = await self.receive()
            self.kill()

    agent = MockedAgentFactory()

    template = Template(body="received body")
    behaviour = RecvBehaviour()
    agent.add_behaviour(behaviour, template)

    await agent.start(auto_register=False)

    await behaviour.join()

    assert behaviour.mailbox_size() == 0
    assert agent.recv_msg is None
    await agent.stop()


async def test_set_get():
    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            self.set("key", "value")
            assert self.get("key") == "value"
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert agent.get("key") == "value"

    await agent.stop()


async def test_multiple_templates():
    class Template1Behaviour(OneShotBehaviour):
        async def run(self):
            self.agent.msg1 = await self.receive(timeout=2)

    class Template2Behaviour(OneShotBehaviour):
        async def run(self):
            self.agent.msg2 = await self.receive(timeout=2)

    class Template3Behaviour(OneShotBehaviour):
        async def run(self):
            self.agent.msg3 = await self.receive(timeout=2)
            self.kill()

    agent = MockedAgentFactory()

    template1 = Template()
    template1.set_metadata("performative", "template1")
    agent.add_behaviour(Template1Behaviour(), template1)

    template2 = Template()
    template2.set_metadata("performative", "template2")
    agent.add_behaviour(Template2Behaviour(), template2)

    template3 = Template()
    template3.set_metadata("performative", "template3")
    behaviour = Template3Behaviour()
    agent.add_behaviour(behaviour, template3)

    msg1 = Message(metadata={"performative": "template1"}).prepare()
    msg2 = Message(metadata={"performative": "template2"}).prepare()
    msg3 = Message(metadata={"performative": "template3"}).prepare()

    await agent.start(auto_register=False)

    agent._message_received(msg1)
    agent._message_received(msg2)
    agent._message_received(msg3)

    await behaviour.join()

    assert agent.msg1.get_metadata("performative") == "template1"
    assert agent.msg2.get_metadata("performative") == "template2"
    assert agent.msg3.get_metadata("performative") == "template3"
    await agent.stop()


async def test_kill_behaviour():
    class TestCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            self.kill()

    agent = MockedAgentFactory()
    behaviour = TestCyclicBehaviour()
    await agent.start(auto_register=False)

    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert behaviour.is_killed()
    assert behaviour.exit_code == 0

    await agent.stop()


async def test_exit_code_from_kill_behaviour():
    class TestCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            self.kill(42)

    agent = MockedAgentFactory()
    behaviour = TestCyclicBehaviour()
    await agent.start(auto_register=False)

    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert behaviour.is_killed()
    assert behaviour.exit_code == 42

    await agent.stop()


async def test_set_exit_code_behaviour():
    class TestCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            self.exit_code = 1024
            await agent.event.wait()
            self.kill()

    agent = MockedAgentFactory()
    behaviour = TestCyclicBehaviour()
    agent.event = asyncio.Event()
    await agent.start(auto_register=False)

    agent.add_behaviour(behaviour)

    with pytest.raises(BehaviourNotFinishedException):
        assert behaviour.exit_code

    agent.event.set()
    await behaviour.join()

    await agent.stop()

    assert behaviour.exit_code == 1024

    assert not agent.is_alive()


async def test_notfinishedexception_behaviour():
    class TestBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.event.wait()

    agent = MockedAgentFactory()
    agent.event = asyncio.Event()
    behaviour = TestBehaviour()
    await agent.start(auto_register=False)

    agent.add_behaviour(behaviour)

    with pytest.raises(BehaviourNotFinishedException):
        assert behaviour.exit_code

    agent.event.set()

    assert behaviour.exit_code == 0

    await agent.stop()

    assert not agent.is_alive()


async def test_cyclic_behaviour():
    class TestCyclicBehaviour(CyclicBehaviour):
        async def run(self):
            self.agent.cycles += 1
            if self.agent.cycles > 2:
                self.kill()

    agent = MockedAgentFactory()
    agent.cycles = 0
    behaviour = TestCyclicBehaviour()
    await agent.start(auto_register=False)

    assert agent.cycles == 0

    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert agent.cycles == 3
    assert behaviour.is_killed()

    await agent.stop()


async def test_oneshot_behaviour():
    class TestOneShotBehaviour(OneShotBehaviour):
        async def run(self):
            self.agent.one_shot_behaviour_executed = True
            self.kill()

    agent = MockedAgentFactory()
    agent.one_shot_behaviour_executed = False
    behaviour = TestOneShotBehaviour()
    agent.add_behaviour(behaviour)

    assert agent.one_shot_behaviour_executed is False

    await agent.start(auto_register=False)

    await behaviour.join()

    assert agent.one_shot_behaviour_executed is True
    await agent.stop()


async def test_periodic_behaviour():
    class TestPeriodicBehaviour(PeriodicBehaviour):
        async def run(self):
            self.agent.periodic_behaviour_execution_counter += 1
            self.kill()

    agent = MockedAgentFactory()
    agent.periodic_behaviour_execution_counter = 0
    behaviour = TestPeriodicBehaviour(0.01)
    agent.add_behaviour(behaviour)

    assert agent.periodic_behaviour_execution_counter == 0

    await agent.start(auto_register=False)

    await behaviour.join()

    assert agent.periodic_behaviour_execution_counter == 1
    await agent.stop()


async def test_periodic_behaviour_period_zero():
    class TestPeriodicBehaviour(PeriodicBehaviour):
        async def run(self):
            self.agent.periodic_behaviour_execution_counter += 1
            self.kill()

    agent = MockedAgentFactory()
    agent.periodic_behaviour_execution_counter = 0
    behaviour = TestPeriodicBehaviour(0)
    agent.add_behaviour(behaviour)

    assert agent.periodic_behaviour_execution_counter == 0

    await agent.start(auto_register=False)

    await behaviour.join()

    assert agent.periodic_behaviour_execution_counter == 1
    await agent.stop()


def test_periodic_behaviour_negative_period():
    class TestPeriodicBehaviour(PeriodicBehaviour):
        async def run(self):
            pass

    with pytest.raises(ValueError):
        TestPeriodicBehaviour(-1)


def test_set_period():
    class TestPeriodicBehaviour(PeriodicBehaviour):
        async def run(self):
            pass

    behaviour = TestPeriodicBehaviour(1)
    assert behaviour.period == datetime.timedelta(seconds=1)

    behaviour.period = 2
    assert behaviour.period == datetime.timedelta(seconds=2)


async def test_periodic_start_at():
    class TestPeriodicBehaviour(PeriodicBehaviour):
        async def run(self):
            self.agent.delay = datetime.datetime.now()
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    start_at = datetime.datetime.now() + datetime.timedelta(seconds=0.01)
    behaviour = TestPeriodicBehaviour(period=0.01, start_at=start_at)

    assert behaviour._next_activation == start_at

    agent.add_behaviour(behaviour)

    await behaviour.join()

    assert agent.delay >= start_at

    await agent.stop()


async def test_timeout_behaviour():
    class TestTimeoutBehaviour(TimeoutBehaviour):
        async def run(self):
            self.agent.delay = datetime.datetime.now()
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    start_at = datetime.datetime.now() + datetime.timedelta(seconds=0.02)
    behaviour = TestTimeoutBehaviour(start_at=start_at)

    assert behaviour._timeout == start_at
    assert not behaviour._timeout_triggered

    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert agent.delay >= start_at
    assert behaviour._timeout_triggered
    assert behaviour.is_done()

    await agent.stop()


async def test_timeout_behaviour_zero():
    class TestTimeoutBehaviour(TimeoutBehaviour):
        async def run(self):
            self.agent.delay = datetime.datetime.now()
            self.kill()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    start_at = datetime.datetime.now() + datetime.timedelta(seconds=0)
    behaviour = TestTimeoutBehaviour(start_at=start_at)

    assert behaviour._timeout == start_at

    assert not behaviour._timeout_triggered

    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert agent.delay >= start_at
    assert behaviour._timeout_triggered
    assert behaviour.is_done()

    await agent.stop()


async def test_fsm_behaviour(fsm):
    agent = MockedAgentFactory()
    agent.sync1_behaviour = asyncio.Event()
    agent.sync2_behaviour = asyncio.Event()
    await agent.start(auto_register=False)

    agent.state = None

    assert len(fsm._transitions) == 2
    assert fsm.current_state == STATE_ONE

    agent.add_behaviour(fsm)
    assert fsm.current_state == STATE_ONE
    assert not fsm.state_one.is_done()
    await wait_for_behaviour_is_killed(fsm.state_one)
    assert agent.state == STATE_ONE
    agent.sync1_behaviour.set()
    await fsm.state_one.join()

    assert fsm.current_state == STATE_TWO
    assert not fsm.state_two.is_done()
    await wait_for_behaviour_is_killed(fsm.state_two)
    assert agent.state == STATE_TWO
    agent.sync2_behaviour.set()
    await fsm.state_two.join()

    assert fsm.current_state == STATE_THREE
    await wait_for_behaviour_is_killed(fsm.state_three)
    assert agent.state == STATE_THREE
    await fsm.state_three.join()

    await agent.stop()


def test_fsm_add_bad_state(fsm):
    with pytest.raises(AttributeError):
        fsm.add_state("BAD_STATE", object())


def test_fsm_is_valid_transition(fsm):
    fsm.is_valid_transition(STATE_ONE, STATE_TWO)


def test_fsm_not_valid_transition(fsm):
    with pytest.raises(NotValidTransition):
        fsm.is_valid_transition(STATE_ONE, STATE_THREE)


def test_fsm_not_valid_state(fsm):
    with pytest.raises(NotValidState):
        fsm.is_valid_transition(STATE_ONE, "BAD_STATE")

    with pytest.raises(NotValidState):
        fsm.is_valid_transition("BAD_STATE", STATE_TWO)


async def test_fsm_bad_state():
    class StateOne(State):
        async def run(self):
            self.set_next_state("BAD_STATE")
            self.kill()

    class StateTwo(State):
        async def run(self):
            pass

    class BadFSMBehaviour(FSMBehaviour):
        async def on_end(self):
            self.kill()

    fsm_ = BadFSMBehaviour()
    state_one = StateOne()
    state_two = StateTwo()
    fsm_.add_state(STATE_ONE, state_one, initial=True)
    fsm_.add_state(STATE_TWO, state_two)
    fsm_.add_transition(STATE_ONE, STATE_TWO)

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    assert fsm_.current_state == STATE_ONE

    agent.add_behaviour(fsm_)

    await state_one.join()
    assert fsm_.current_state == STATE_ONE

    await fsm_.join()
    assert fsm_.is_killed()

    await agent.stop()


async def test_fsm_bad_transition():
    class StateOne(State):
        async def run(self):
            self.set_next_state(STATE_THREE)
            self.kill()

    class StateTwo(State):
        async def run(self):
            pass

    class StateThree(State):
        async def run(self):
            pass

    class BadFSMBehaviour(FSMBehaviour):
        async def on_end(self):
            self.kill()

    fsm_ = BadFSMBehaviour()
    state_one = StateOne()
    state_two = StateTwo()
    state_three = StateThree()
    fsm_.add_state(STATE_ONE, state_one, initial=True)
    fsm_.add_state(STATE_TWO, state_two)
    fsm_.add_state(STATE_THREE, state_three)
    fsm_.add_transition(STATE_ONE, STATE_TWO)
    fsm_.add_transition(STATE_TWO, STATE_THREE)

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    assert fsm_.current_state == STATE_ONE

    agent.add_behaviour(fsm_)

    await state_one.join()
    assert fsm_.current_state == STATE_ONE

    await fsm_.join()
    assert fsm_.is_killed()

    await agent.stop()


def test_fsm_two_initials():
    class StateOne(State):
        async def run(self):
            self.set_next_state(STATE_THREE)
            self.kill()

    class StateTwo(State):
        async def run(self):
            pass

    class TestFSMBehaviour(FSMBehaviour):
        async def on_end(self):
            self.kill()

    fsm_ = TestFSMBehaviour()
    state_one = StateOne()
    state_two = StateTwo()
    fsm_.add_state(STATE_ONE, state_one, initial=True)
    fsm_.add_state(STATE_TWO, state_two, initial=True)

    assert fsm_.current_state == STATE_TWO


async def test_fsm_fail_on_start():
    class StateOne(State):
        async def on_start(self):
            raise Exception

        async def run(self):
            pass

    fsm_ = FSMBehaviour()
    state_one = StateOne()
    fsm_.add_state(STATE_ONE, state_one, initial=True)

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.add_behaviour(fsm_)

    await fsm_.join()

    assert fsm_.is_killed()

    assert type(fsm_.exit_code) == Exception

    await agent.stop()


async def test_fsm_fail_on_run():
    class StateOne(State):
        async def run(self):
            raise Exception

    fsm_ = FSMBehaviour()
    state_one = StateOne()
    fsm_.add_state(STATE_ONE, state_one, initial=True)

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.add_behaviour(fsm_)

    await fsm_.join()

    assert fsm_.is_killed()

    assert type(fsm_.exit_code) == Exception

    await agent.stop()


async def test_fsm_fail_on_end():
    class StateOne(State):
        async def run(self):
            pass

        async def on_end(self):
            raise Exception

    fsm_ = FSMBehaviour()
    state_one = StateOne()
    fsm_.add_state(STATE_ONE, state_one, initial=True)

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.add_behaviour(fsm_)

    await fsm_.join()

    assert fsm_.is_killed()

    assert type(fsm_.exit_code) == Exception

    await agent.stop()


def test_to_graphviz(fsm):
    assert (
        fsm.to_graphviz()
        == "digraph finite_state_machine { rankdir=LR; node [fixedsize=true];STATE_ONE -> STATE_TWO;STATE_TWO -> STATE_THREE;}"
    )


async def test_join():
    class WaitBehav(OneShotBehaviour):
        async def run(self):
            for i in range(100):
                self.agent.i = i
                await asyncio.sleep(0)

    agent = MockedAgentFactory()
    agent.i = None
    behaviour = WaitBehav()

    agent.add_behaviour(behaviour)
    await agent.start(auto_register=False)

    await behaviour.join(timeout=None)

    assert behaviour.is_done()
    assert agent.i == 99

    await agent.stop()


async def test_join_with_timeout():
    class WaitBehav(OneShotBehaviour):
        async def run(self):
            await asyncio.sleep(100)

    agent = MockedAgentFactory()
    agent.i = None
    behaviour = WaitBehav()

    agent.add_behaviour(behaviour)
    await agent.start(auto_register=False)

    with pytest.raises(TimeoutError):
        await behaviour.join(timeout=0.01)

    assert not behaviour.is_done()

    await agent.stop()


async def test_join_with_long_timeout():
    class WaitBehav(OneShotBehaviour):
        async def run(self):
            await asyncio.sleep(0)

    agent = MockedAgentFactory()
    agent.i = None
    behaviour = WaitBehav()

    agent.add_behaviour(behaviour)
    await agent.start(auto_register=False)

    await behaviour.join(timeout=100)

    assert behaviour.is_done()

    await agent.stop()


async def test_join_inside_behaviour():
    class Behav1(OneShotBehaviour):
        async def run(self):
            class Behav2(OneShotBehaviour):
                async def run(self):
                    self.agent.behav2 = True

            behav2 = Behav2()
            self.agent.add_behaviour(behav2)
            await behav2.join()
            self.agent.behav1 = True

    agent = MockedAgentFactory()
    agent.behav1 = False
    agent.behav2 = False

    behav1 = Behav1()
    agent.add_behaviour(behav1)

    await agent.start(auto_register=False)

    await behav1.join()

    assert agent.behav1
    assert agent.behav2

    await agent.stop()


async def test_join_inside_behaviour_with_timeout():
    class Behav1(OneShotBehaviour):
        async def run(self):
            class Behav2(OneShotBehaviour):
                async def run(self):
                    await asyncio.sleep(1)

            behav2 = Behav2()
            self.agent.add_behaviour(behav2)
            with pytest.raises(TimeoutError):
                await behav2.join(timeout=0.001)
            self.agent.behav1 = True

    agent = MockedAgentFactory()
    agent.behav1 = False

    behav1 = Behav1()
    agent.add_behaviour(behav1)

    await agent.start(auto_register=False)

    await behav1.join()

    assert agent.behav1

    await agent.stop()


async def test_behaviour_at_end():
    class FinalBehav(OneShotBehaviour):
        async def run(self):
            self.agent.value = 2000

    class StopAgent(Agent):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._async_connect = AsyncMock()
            self._async_register = AsyncMock()
            self.conn_coro = Mock()
            self.conn_coro.__aexit__ = AsyncMock()
            self.stream = Mock()
            self.value = 1000

        async def stop(self):
            behav = FinalBehav()
            self.add_behaviour(behav)
            await behav.join()
            return await super().stop()

    agent = StopAgent("fakejid@fakeserver", "fakepassword")

    await agent.start(auto_register=False)

    assert agent.value == 1000
    await agent.stop()

    assert agent.value == 2000


async def test_two_behaviours_at_end():
    class FinalBehav2(OneShotBehaviour):
        async def run(self):
            self.agent.value = 2000

    class FinalBehav1(OneShotBehaviour):
        async def run(self):
            behav = FinalBehav2()
            self.agent.add_behaviour(behav)
            await behav.join()

    class StopAgent(Agent):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._async_connect = AsyncMock()
            self._async_register = AsyncMock()
            self.conn_coro = Mock()
            self.conn_coro.__aexit__ = AsyncMock()
            self.stream = Mock()
            self.value = 1000

        async def stop(self):
            behav = FinalBehav1()
            self.add_behaviour(behav)
            await behav.join()
            return await super().stop()

    agent = StopAgent("fakejid@fakeserver", "fakepassword")

    await agent.start(auto_register=False)

    assert agent.value == 1000
    await agent.stop()

    assert agent.value == 2000


def test_get_state(fsm):
    assert isinstance(fsm.get_state(STATE_ONE), StateOne)
    assert isinstance(fsm.get_state(STATE_TWO), StateTwo)
    assert isinstance(fsm.get_state(STATE_THREE), StateThree)



# -*- coding: utf-8 -*-
import asyncio
import logging
from typing import Union, List

from . import agent
from . import behaviour
from . import message
from . import template
from .agent import AgentType
from .container import run_container as run  # noqa: F401

__author__ = """Javi Palanca"""
__email__ = "jpalanca@gmail.com"
__version__ = "4.0.0rc2"

__all__ = ["agent", "behaviour", "message", "template"]

logger = logging.getLogger("SPADE")


async def wait_until_finished(agents: Union[List[AgentType], AgentType]) -> None:
    if not isinstance(agents, list):
        agents = [agents]
    try:
        while any([ag.is_alive() for ag in agents]):
            await asyncio.sleep(1)
    except KeyboardInterrupt:
        logger.warning("Keyboard interrupt received. Stopping SPADE...")
    for ag in agents:
        logger.info(f"Stopping agent {ag.jid}")
        await ag.stop()



import asyncio
import logging
from asyncio import Task
from hashlib import md5
from typing import Coroutine, Optional, Type, Any, List, TypeVar

from slixmpp import JID
from slixmpp import Message as slixmppMessage

from .behaviour import BehaviourType, FSMBehaviour, CyclicBehaviour
from .container import Container
from .message import Message
from .presence import PresenceManager
from .template import Template
from .trace import TraceStore
from .web import WebApp
from .xmpp_client import XMPPClient

logger = logging.getLogger("spade.Agent")

AgentType = TypeVar("AgentType", bound="Agent")


class AuthenticationFailure(Exception):
    """ """

    pass


class DisconnectedException(Exception):
    """ """

    pass


class Agent(object):
    def __init__(
        self, jid: str, password: str, port: int = 5222, verify_security: bool = False
    ):
        """
        Creates an agent

        Args:
          jid (str): The identifier of the agent in the form username@server
          password (str): The password to connect to the server
          verify_security (bool): Weather to verify or not the SSL certificates
        """
        self.jid = JID(jid)
        self.password = password
        self.xmpp_port = port
        self.verify_security = verify_security

        self.behaviours = []
        self._values = {}

        self.client = None
        self.presence = None
        self.loop = None

        self.container = Container()
        self.container.register(self)

        self.loop = self.container.loop
        asyncio.set_event_loop(self.loop)

        # Web service
        self.web = WebApp(agent=self)

        self.traces = TraceStore(size=1000)

        self._alive = asyncio.Event()

    def set_loop(self, loop) -> None:
        self.loop = loop

    def set_container(self, container: Container) -> None:
        """
        Sets the container to which the agent is attached

        Args:
            container (spade.container.Container): the container to be attached to
        """
        self.container = container

    async def start(self, auto_register: bool = True) -> None:
        """
        Starts this agent.

        Args:
            auto_register (bool): register the agent in the server (Default value = True)

        Returns:
            None
        """
        return await self._async_start(auto_register=auto_register)

    async def _async_start(self, auto_register: bool = True) -> None:
        """
        Starts the agent from a coroutine. This fires some actions:

            * if auto_register: register the agent in the server
            * runs the event loop
            * connects the agent to the server
            * runs the registered behaviours

        Args:
          auto_register (bool, optional): register the agent in the server (Default value = True)

        """
        await self._hook_plugin_before_connection()

        self.client = XMPPClient(
            self.jid, self.password, self.verify_security, auto_register
        )
        # Presence service
        self.presence = PresenceManager(agent=self, approve_all=False)

        await self._async_connect()

        await self._hook_plugin_after_connection()

        await self.setup()
        self._alive.set()
        for behaviour in self.behaviours:
            if not behaviour.is_running:
                behaviour.set_agent(self)
                if issubclass(type(behaviour), FSMBehaviour):
                    for _, state in behaviour.get_states().items():
                        state.set_agent(self)
                behaviour.start()

    async def _hook_plugin_before_connection(self) -> None:
        """
        Overload this method to hook a plugin before connection is done
        """
        pass

    async def _hook_plugin_after_connection(self) -> None:
        """
        Overload this method to hook a plugin after connection is done
        """
        pass

    async def _async_connect(self) -> None:  # pragma: no cover
        """connect and authenticate to the XMPP server. Async mode."""

        self.client.connected_event = asyncio.Event()
        self.client.disconnected_event = asyncio.Event()
        self.client.failed_auth_event = asyncio.Event()

        connected_task = asyncio.create_task(
            self.client.connected_event.wait(), name="connected"
        )
        disconnected_task = asyncio.create_task(
            self.client.disconnected_event.wait(), name="disconnected"
        )
        failed_auth_task = asyncio.create_task(
            self.client.failed_auth_event.wait(), name="failed_auth"
        )

        self.client.add_event_handler(
            "session_start", lambda _: self.client.connected_event.set()
        )
        self.client.add_event_handler(
            "disconnected", lambda _: self.client.disconnected_event.set()
        )
        self.client.add_event_handler(
            "failed_all_auth", lambda _: self.client.failed_auth_event.set()
        )
        self.client.add_event_handler("message", self._message_received)

        self.client.connect(self.jid.host, self.xmpp_port)

        done, pending = await asyncio.wait(
            [connected_task, disconnected_task, failed_auth_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        for task in done:
            await task

            if task.get_name() == "failed_auth":
                raise AuthenticationFailure(
                    "Could not authenticate the agent. Check user and password or use auto_register=True"
                )
            elif task.get_name() == "disconnected":
                raise DisconnectedException(
                    "Error during the connection with the server"
                )

        logger.info(f"Agent {str(self.jid)} connected and authenticated.")

    async def setup(self) -> None:
        """
        Setup agent before startup.
        This coroutine may be overloaded.
        """
        await asyncio.sleep(0)

    @property
    def name(self) -> str:
        """
        Returns the name of the agent (the string before the '@')

        Returns:
            str: the name of the agent (the string before the '@')
        """
        return self.jid.node

    @property
    def avatar(self) -> str:
        """
        Generates a unique avatar for the agent based on its JID.
        Uses Gravatar service with MonsterID option.

        Returns:
          str: the url of the agent's avatar

        """
        return self.build_avatar_url(self.jid.bare)

    @staticmethod
    def build_avatar_url(jid: str) -> str:
        """
        Static method to build a gravatar url with the agent's JID

        Args:
          jid (aioxmpp.JID): an XMPP identifier

        Returns:
          str: a URL for the gravatar

        """
        digest = md5(str(jid).encode("utf-8")).hexdigest()
        return "http://www.gravatar.com/avatar/{md5}?d=monsterid".format(md5=digest)

    def submit(self, coro: Coroutine) -> Task:
        """
        Runs a coroutine in the event loop of the agent.
        this call is not blocking.

        Args:
          coro (Coroutine): the coroutine to be run

        Returns:
            asyncio.Task: the Task assigned to the coroutine execution

        """
        return asyncio.create_task(coro)

    def add_behaviour(
        self, behaviour: BehaviourType, template: Optional[Template] = None
    ) -> None:
        """
        Adds and starts a behaviour to the agent.
        If template is not None it is used to match
        new messages and deliver them to the behaviour.

        Args:
          behaviour (Type[spade.behaviour.CyclicBehaviour]): the behaviour to be started
          template (spade.template.Template, optional): the template to match messages with (Default value = None)

        """
        behaviour.set_agent(agent=self)
        if issubclass(type(behaviour), FSMBehaviour):
            for _, state in behaviour.get_states().items():
                state.set_agent(self)
        behaviour.set_template(template)
        self.behaviours.append(behaviour)
        if self.is_alive():
            behaviour.start()

    def remove_behaviour(self, behaviour: Type[CyclicBehaviour]) -> None:
        """
        Removes a behaviour from the agent.
        The behaviour is first killed.

        Args:
          behaviour (Type[spade.behaviour.CyclicBehaviour]): the behaviour instance to be removed

        """
        if not self.has_behaviour(behaviour):
            raise ValueError("This behaviour is not registered")
        index = self.behaviours.index(behaviour)
        self.behaviours[index].kill()
        self.behaviours.pop(index)

    def has_behaviour(self, behaviour: Type[CyclicBehaviour]) -> bool:
        """
        Checks if a behaviour is added to an agent.

        Args:
          behaviour (Type[spade.behaviour.CyclicBehaviour]): the behaviour instance to check

        Returns:
          bool: a boolean that indicates wether the behaviour is inside the agent.

        """
        return behaviour in self.behaviours

    async def stop(self) -> None:
        """
        Stops this agent.
        """
        return await self._async_stop()

    async def _async_stop(self) -> None:
        """Stops an agent and kills all its behaviours."""
        if self.presence:
            self.presence.set_unavailable()
        for behav in self.behaviours:
            behav.kill()
        if self.web.is_started():
            await self.web.runner.cleanup()

        if self.is_alive():
            # Disconnect from XMPP server
            await self.client.disconnect()
            logger.info("Client disconnected.")

        self._alive.clear()

    def is_alive(self) -> bool:
        """
        Checks if the agent is alive.

        Returns:
          bool: wheter the agent is alive or not

        """
        return self._alive.is_set()

    def set(self, name: str, value: Any):
        """
        Stores a knowledge item in the agent knowledge base.

        Args:
          name (str): name of the item
          value (object): value of the item

        """
        self._values[name] = value

    def get(self, name: str) -> Any:
        """
        Recovers a knowledge item from the agent's knowledge base.

        Args:
          name(str): name of the item

        Returns:
          object: the object retrieved or None

        """
        if name in self._values:
            return self._values[name]
        else:
            return None

    def _message_received(self, msg: slixmppMessage) -> List[Task]:
        """
        Callback run when an XMPP Message is reveived.
        This callback delivers the message to every behaviour
        that is waiting for it. First, the aioxmpp.Message is
        converted to spade.message.Message

        Args:
          msg (slixmpp.Message): the message just received.

        Returns:
            list(asyncio.Future): a list of futures of the appended of the message at each matched behaviour.

        """

        msg = Message.from_node(msg)
        return self.dispatch(msg)

    def dispatch(self, msg: Message) -> List[Task]:
        """
        Dispatch the message to every behaviour that is waiting for
        it using their templates match.

        Args:
          msg (spade.message.Message): the message to dispatch.

        Returns:
            list(asyncio.Future): a list of tasks for each message queuing in each matching behavior.

        """
        logger.debug(f"Got message: {msg}")
        tasks = []
        matched = False
        for behaviour in (x for x in self.behaviours if x.match(msg)):
            tasks.append(self.submit(behaviour.enqueue(msg)))
            logger.debug(f"Message enqueued to behaviour: {behaviour}")
            self.traces.append(msg, category=str(behaviour))
            matched = True
        if not matched:
            logger.warning(f"No behaviour matched for message: {msg}")
            self.traces.append(msg)
        return tasks



import asyncio
import collections
import logging
import time
import traceback
from abc import ABCMeta, abstractmethod
from asyncio import CancelledError
from datetime import timedelta, datetime
from threading import Event
from typing import Any, Optional, Dict, TypeVar

from .message import Message
from .template import Template

now = datetime.now

logger = logging.getLogger("spade.behaviour")

BehaviourType = TypeVar("BehaviourType", bound="CyclicBehaviour")


class BehaviourNotFinishedException(Exception):
    """ """

    pass


class NotValidState(Exception):
    """ """

    pass


class NotValidTransition(Exception):
    """ """

    pass


class CyclicBehaviour(object, metaclass=ABCMeta):
    """This behaviour is executed cyclically until it is stopped."""

    def __init__(self):
        self.agent = None
        self.template = None
        self._force_kill = Event()
        self._is_done = None
        self._exit_code = 0
        self.presence = None
        self.web = None
        self.is_running = False

        self.queue = None

    def set_agent(self, agent) -> None:
        """
        Links behaviour with its owner agent

        Args:
          agent (spade.agent.Agent): the agent who owns the behaviour

        """
        self.agent = agent
        asyncio.set_event_loop(self.agent.loop)
        self._is_done = asyncio.Event()
        self._is_done.set()
        self.queue = asyncio.Queue()
        self.presence = agent.presence
        self.web = agent.web

    def set_template(self, template: Template) -> None:
        """
        Sets the template that is used to match incoming
        messages with this behaviour.

        Args:
          template (spade.template.Template): the template to match with

        """
        self.template = template

    def match(self, message: Message) -> bool:
        """
        Matches a message with the behaviour's template

        Args:
          message(spade.message.Message): the message to match with

        Returns:
          bool: wheter the messaged matches or not

        """
        if self.template:
            return self.template.match(message)
        return True

    def set(self, name: str, value: Any) -> None:
        """
        Stores a knowledge item in the agent knowledge base.

        Args:
          name (str): name of the item
          value (Any): value of the item

        """
        self.agent.set(name, value)

    def get(self, name: str) -> Any:
        """
        Recovers a knowledge item from the agent's knowledge base.

        Args:
          name (str): name of the item

        Returns:
          Any: the object retrieved or None

        """
        return self.agent.get(name)

    def start(self) -> None:
        """starts behaviour in the event loop"""
        self.agent.submit(self._start())
        self.is_running = True

    async def _start(self) -> None:
        """
        Start coroutine. runs on_start coroutine and then
        runs the _step coroutine where the body of the behaviour
        is called.
        """
        await self.agent._alive.wait()
        try:
            await self.on_start()
        except Exception as e:
            logger.error(
                "Exception running on_start in behaviour {}: {}".format(self, e)
            )
            self.kill(exit_code=e)
        await self._step()
        self._is_done.clear()

    def kill(self, exit_code: Optional[Any] = None) -> None:
        """
        Stops the behaviour

        Args:
          exit_code (object, optional): the exit code of the behaviour (Default value = None)

        """
        self._force_kill.set()
        if exit_code is not None:
            self._exit_code = exit_code
        logger.info("Killing behavior {0} with exit code: {1}".format(self, exit_code))

    def is_killed(self) -> bool:
        """
        Checks if the behaviour was killed by means of the kill() method.

        Returns:
          bool: whether the behaviour is killed or not

        """
        return self._force_kill.is_set()

    @property
    def exit_code(self) -> Any:
        """
        Returns the exit_code of the behaviour.
        It only works when the behaviour is done or killed,
        otherwise it raises an exception.

        Returns:
          object: the exit code of the behaviour

        Raises:
            BehaviourNotFinishedException: if the behaviour is not yet finished

        """
        if self._done() or self.is_killed():
            return self._exit_code
        else:
            raise BehaviourNotFinishedException

    @exit_code.setter
    def exit_code(self, value: Any) -> None:
        """
        Sets a new exit code to the behaviour.

        Args:
          value (object): the new exit code

        """
        self._exit_code = value

    def _done(self) -> bool:
        """
        Returns True if the behaviour has finished
        else returns False

        Returns:
          bool: whether the behaviour is finished or not

        """
        return False

    def is_done(self) -> bool:
        """
        Check if the behaviour is finished

        Returns:
             bool: whether the behaviour is finished or not
        """
        return not self._is_done.is_set()

    async def join(self, timeout: Optional[float] = None) -> None:
        """
        Wait for the behaviour to complete

        Args:
            timeout (Optional[float]): an optional timeout to wait to join (if None, the join is blocking)

        Returns:
            None

        Raises:
            TimeoutError: if the timeout is reached
        """

        return await self._async_join(timeout=timeout)

    async def _async_join(self, timeout: Optional[float]) -> None:
        """
        Coroutine to wait until a behaviour is finished

        Args:
            timeout (Optional[float]): an optional timeout to wait to join

        Raises:
            TimeoutError: fi the timeout is reached
        """
        t_start = time.time()
        while not self.is_done():
            await asyncio.sleep(0.001)
            t = time.time()
            if timeout is not None and t - t_start > timeout:
                raise TimeoutError

    async def on_start(self) -> None:
        """
        Coroutine called before the behaviour is started.
        """
        pass

    async def on_end(self) -> None:
        """
        Coroutine called after the behaviour is done or killed.
        """
        pass

    @abstractmethod
    async def run(self) -> None:
        """
        Body of the behaviour.
        To be implemented by user.
        """
        raise NotImplementedError  # pragma: no cover

    async def _run(self) -> None:
        """
        Function to be overload by more complex behaviours.
        In other case it just calls run() coroutine.
        """
        await self.run()

    async def _step(self) -> None:
        """
        Main loop of the behaviour.
        checks whether behaviour is done or killed,
        ortherwise it calls run() coroutine.
        """
        cancelled = False
        while not self._done() and not self.is_killed():
            try:
                await self._run()
                await asyncio.sleep(0)  # relinquish cpu
            except CancelledError:  # pragma: no cover
                logger.debug("Behaviour {} cancelled".format(self))
                cancelled = True
            except Exception as e:
                logger.error(
                    "Exception running behaviour {behav}: {exc}".format(
                        behav=self, exc=e
                    )
                )
                logger.error(traceback.format_exc())
                self.kill(exit_code=e)
        try:
            if not cancelled:
                await self.on_end()
        except Exception as e:
            logger.error("Exception running on_end in behaviour {}: {}".format(self, e))
            self.kill(exit_code=e)
        self.is_running = False
        self.agent.remove_behaviour(self)

    async def enqueue(self, message: Message) -> None:
        """
        Enqueues a message in the behaviour's mailbox

        Args:
            message (spade.message.Message): the message to be enqueued
        """
        asyncio.create_task(self.queue.put(message))

    def mailbox_size(self) -> int:
        """
        Checks if there is a message in the mailbox

        Returns:
          int: the number of messages in the mailbox

        """
        return self.queue.qsize()

    async def send(self, msg: Message) -> None:
        """
        Sends a message.

        Args:
            msg (spade.message.Message): the message to be sent.
        """
        if not msg.sender:
            msg.sender = str(self.agent.jid)
            logger.debug(f"Adding agent's jid as sender to message: {msg}")
        await self.agent.container.send(msg, self)
        msg.sent = True
        self.agent.traces.append(msg, category=str(self))

    async def _xmpp_send(self, msg: Message) -> None:
        slixmpp_msg = msg.prepare()
        await self.agent.client.send(slixmpp_msg)

    async def receive(self, timeout: Optional[float] = None) -> Optional[Message]:
        """
        Receives a message for this behaviour.
        If timeout is not None it returns the message or "None"
        after timeout is done.

        Args:
            timeout (float, optional): number of seconds until return

        Returns:
            spade.message.Message: a Message or None
        """
        if timeout:
            coro = self.queue.get()
            try:
                msg = await asyncio.wait_for(coro, timeout=timeout)
            except asyncio.TimeoutError:
                msg = None
        else:
            try:
                msg = self.queue.get_nowait()
            except asyncio.QueueEmpty:
                msg = None
        return msg

    def __str__(self) -> str:
        return "{}/{}".format(
            "/".join(base.__name__ for base in self.__class__.__bases__),
            self.__class__.__name__,
        )


class OneShotBehaviour(CyclicBehaviour, metaclass=ABCMeta):
    """This behaviour is only executed once"""

    def __init__(self):
        super().__init__()
        self._already_executed = False

    def _done(self) -> bool:
        """ """
        if not self._already_executed:
            self._already_executed = True
            return False
        return True


class PeriodicBehaviour(CyclicBehaviour, metaclass=ABCMeta):
    """This behaviour is executed periodically with an interval"""

    def __init__(self, period: float, start_at: Optional[datetime] = None):
        """
        Creates a periodic behaviour.

        Args:
            period (float): interval of the behaviour in seconds
            start_at (datetime.datetime): whether to start the behaviour with an offset
        """
        super().__init__()
        self._period = None
        self.period = period

        if start_at:
            self._next_activation = start_at
        else:
            self._next_activation = now()

    @property
    def period(self) -> timedelta:
        """Get the period."""
        return self._period

    @period.setter
    def period(self, value: float) -> None:
        """
        Set the period.

        Args:
          value (float): seconds
        """
        if value < 0:
            raise ValueError("Period must be greater or equal than zero.")
        self._period = timedelta(seconds=value)

    async def _run(self) -> None:
        if now() >= self._next_activation:
            logger.debug(f"Periodic behaviour activated: {self}")
            await self.run()
            if self.period <= timedelta(seconds=0):
                self._next_activation = now()
            else:
                while self._next_activation <= now():
                    self._next_activation += self._period
        else:
            seconds = (self._next_activation - now()).total_seconds()
            if seconds > 0:
                logger.debug(
                    f"Periodic behaviour going to sleep for {seconds} seconds: {self}"
                )
                await asyncio.sleep(seconds)


class TimeoutBehaviour(OneShotBehaviour, metaclass=ABCMeta):
    """This behaviour is executed once at after specified datetime"""

    def __init__(self, start_at):
        """
        Creates a timeout behaviour, which is run at start_at

        Args:
            start_at (datetime.datetime): when to start the behaviour
        """
        super().__init__()

        self._timeout = start_at
        self._timeout_triggered = False

    async def _run(self) -> None:
        if now() >= self._timeout:
            logger.debug(f"Timeout behaviour activated: {self}")
            await self.run()
            self._timeout_triggered = True
        else:
            seconds = (self._timeout - now()).total_seconds()
            if seconds > 0:
                logger.debug(
                    f"Timeout behaviour going to sleep for {seconds} seconds: {self}"
                )
                start_time = time.monotonic()
                await asyncio.sleep(seconds)
                elapsed = time.monotonic() - start_time
                if elapsed >= seconds:
                    await self.run()
                    self._timeout_triggered = True

    def _done(self) -> bool:
        """ """
        return self._timeout_triggered


class State(OneShotBehaviour, metaclass=ABCMeta):
    """A state of a FSMBehaviour is a OneShotBehaviour"""

    def __init__(self):
        super().__init__()
        self.next_state = None

    def set_next_state(self, state_name: str) -> None:
        """
        Set the state to transition to when this state is finished.
        state_name must be a valid state and the transition must be registered.
        If set_next_state is not called then current state is a final state.

        Args:
          state_name (str): the name of the state to transition to

        """
        self.next_state = state_name


class FSMBehaviour(CyclicBehaviour):
    """A behaviour composed of states (oneshotbehaviours) that may transition from one state to another."""

    def __init__(self):
        super().__init__()
        self._states: Dict[str, State] = {}
        self._transitions = collections.defaultdict(list)
        self.current_state: Optional[str] = None
        self.setup()

    def setup(self) -> None:
        """ """
        pass

    def add_state(self, name: str, state: State, initial: bool = False) -> None:
        """Adds a new state to the FSM.

        Args:
          name (str): the name of the state, which is used as its identifier.
          state (spade.behaviour.State): The state class
          initial (bool, optional): wether the state is the initial state or not. (Only one initial state is allowed) (Default value = False)

        """
        if not issubclass(state.__class__, State):
            raise AttributeError("state must be subclass of spade.behaviour.State")
        self._states[name] = state
        if initial:
            self.current_state = name

    def get_state(self, name) -> State:
        return self._states[name]

    def get_states(self) -> Dict[str, State]:
        return self._states

    def add_transition(self, source: str, dest: str) -> None:
        """Adds a transition from one state to another.

        Args:
          source (str): the name of the state from where the transition starts
          dest (str): the name of the state where the transition ends

        """
        self._transitions[source].append(dest)

    def is_valid_transition(self, source: str, dest: str) -> bool:
        """
        Checks if a transitions is registered in the FSM

        Args:
          source (str): the source state name
          dest (str): the destination state name

        Returns:
          bool: wether the transition is valid or not

        """
        if dest not in self._states or source not in self._states:
            raise NotValidState
        elif dest not in self._transitions[source]:
            raise NotValidTransition
        return True

    async def _run(self) -> None:
        behaviour = self._states[self.current_state]
        behaviour.set_agent(self.agent)
        behaviour.receive = self.receive
        logger.info(f"FSM running state {self.current_state}")
        try:
            await behaviour.on_start()
        except Exception as e:
            logger.error("Exception running on_start in state {}: {}".format(self, e))
            self.kill(exit_code=e)
        try:
            await behaviour.run()
        except Exception as e:
            logger.error("Exception running state {}: {}".format(self, e))
            self.kill(exit_code=e)
        try:
            await behaviour.on_end()
        except Exception as e:
            logger.error("Exception running on_start in state {}: {}".format(self, e))
            self.kill(exit_code=e)

        dest = behaviour.next_state
        behaviour._is_done.clear()

        if dest:
            try:
                if self.is_valid_transition(self.current_state, dest):
                    logger.info(f"FSM transiting from {self.current_state} to {dest}.")
                    self.current_state = dest
            except NotValidState as e:
                logger.error(
                    f"FSM could not transitate to state {dest}. That state does not exist."
                )
                self.kill(exit_code=e)
            except NotValidTransition as e:
                logger.error(
                    f"FSM could not transitate to state {dest}. That transition is not registered."
                )
                self.kill(exit_code=e)
        else:
            logger.info(
                "FSM arrived to a final state (no transitions found). Killing FSM."
            )
            self.kill()

    async def run(self) -> None:
        """
        In this kind of behaviour there is no need to overload run.
        The run methods to be overloaded are in the State class.
        """
        raise RuntimeError  # pragma: no cover

    def to_graphviz(self) -> str:
        """
        Converts the FSM behaviour structure to Graphviz syntax

        Returns:
          str: the graph in Graphviz syntax

        """
        graph = "digraph finite_state_machine { rankdir=LR; node [fixedsize=true];"
        for origin, dest in self._transitions.items():
            origin = origin.replace(" ", "_")
            for d in dest:
                d = d.replace(" ", "_")
                graph += "{0} -> {1};".format(origin, d)
        graph += "}"
        return graph



import sys
import click
import spade
import socket
from pyjabber.server import Server


def check_port_in_use(port, host="0.0.0.0"):
    """Checks if a port is in use"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        result = s.connect_ex((host, port))
        return result == 0


def create_cli():
    """Factory function to create the CLI"""

    @click.group()
    def cli():
        """SPADE command line tool"""
        pass

    @cli.command()
    @click.option("--host", default="0.0.0.0", help="Server host")
    @click.option("--client_port", default=5222, help="Client port")
    @click.option("--server_port", default=5269, help="Server port")
    @click.option("--debug", is_flag=True, default=False, help="Enables debug mode")
    def run(host, client_port, server_port, debug):
        """Launch an XMPP server"""
        if check_port_in_use(client_port, host):
            click.echo(f"Error: The port {client_port} is already in use.")
            sys.exit(1)
        if check_port_in_use(server_port, host):
            click.echo(f"Error: The port {server_port} is already in use.")
            sys.exit(1)

        server = Server(
            host=host,
            client_port=client_port,
            server_port=server_port,
            connection_timeout=600,
        )
        server.start(debug=debug)

    @cli.command()
    def version():
        """Displays the SPADE version"""
        click.echo(spade.__version__)

    return cli


# Create a single instance of the CLI
cli = create_cli()

if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import asyncio

from testfixtures import LogCapture
from unittest.mock import AsyncMock, Mock

from slixmpp import Message as slixmppMessage

import spade.xmpp_client
from spade.agent import Agent
from spade.behaviour import OneShotBehaviour
from spade.message import Message
from spade.template import Template
from .factories import MockedAgentFactory


async def test_create_agent(mocker):
    agent = MockedAgentFactory()

    assert agent.is_alive() is False

    await agent.start(auto_register=False)

    agent._async_connect.assert_called_once()

    await agent.stop()

    assert agent.is_alive() is False


def test_name():
    agent = MockedAgentFactory(jid="john@fake.server")
    assert agent.name == "john"


def test_avatar():
    agent = MockedAgentFactory(jid="test_avatar@fake.server")
    assert (
        agent.avatar
        == "http://www.gravatar.com/avatar/c23af06d68b5e37d3589e7a8748e7b6c?d=monsterid"
    )


async def test_setup():
    agent = MockedAgentFactory()
    agent.setup = AsyncMock()

    await agent.start()

    agent.setup.assert_called_once()

    await agent.stop()


def test_set_get():
    agent = MockedAgentFactory()
    agent.set("KB_name", "KB_value")
    assert agent.get("KB_name") == "KB_value"


def test_get_none():
    agent = MockedAgentFactory()
    assert agent.get("KB_name_unknown") is None


async def test_client():
    agent = MockedAgentFactory()
    assert agent.client is None

    await agent.start(auto_register=False)

    assert type(agent.client) == spade.xmpp_client.XMPPClient

    await agent.stop()

#
# async def test_register():
#     agent = MockedAgentFactory()
#     agent.register = Mock()
#
#     await agent.start()
#
#     assert len(agent._async_register.mock_calls) == 1
#
#     await agent.stop()


def test_receive_without_behaviours():
    agent = MockedAgentFactory()
    slimsg = slixmppMessage()
    slimsg.chat()
    msg = Message.from_node(slimsg)

    assert agent.traces.len() == 0

    with LogCapture() as log:
        agent._message_received(slimsg)
        log.check_present(
            ("spade.Agent", "WARNING", f"No behaviour matched for message: {msg}")
        )

    assert agent.traces.len() == 1
    assert msg in agent.traces.store[0]


async def test_create_agent_from_another_agent():
    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent._done = True
            self.kill()

    class CreateBehav(OneShotBehaviour):
        async def run(self):
            self.agent.agent2 = MockedAgentFactory()
            self.agent.agent2._done = False
            self.agent.dummy_behav = DummyBehav()
            self.agent.agent2.add_behaviour(self.agent.dummy_behav)
            await self.agent.agent2.start(auto_register=False)
            self.kill()

    agent1 = MockedAgentFactory()
    agent1.agent2 = None
    create_behav = CreateBehav()
    agent1.add_behaviour(create_behav)

    await agent1.start(auto_register=False)

    await create_behav.join()

    assert agent1.agent2._done

    await agent1.stop()


async def test_create_agent_from_another_agent_from_setup():
    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent._done = True
            self.kill()

    class SetupAgent(Agent):
        async def setup(self):
            self.agent2 = MockedAgentFactory()
            self.agent2._done = False
            self.agent2.dummy_behav = DummyBehav()
            self.agent2.add_behaviour(self.agent2.dummy_behav)
            await self.agent2.start(auto_register=False)

    agent1 = SetupAgent("fake@host", "secret")
    agent1._async_connect = AsyncMock()
    agent1._async_register = AsyncMock()
    agent1.conn_coro = Mock()
    agent1.conn_coro.__aexit__ = AsyncMock()
    agent1.stream = Mock()

    agent1.agent2 = None

    await agent1.start(auto_register=False)
    assert agent1.is_alive()

    await agent1.agent2.dummy_behav.join()

    assert agent1.agent2.is_alive()
    assert agent1.agent2._done

    await agent1.agent2.stop()
    await agent1.stop()


async def test_submit_send():
    agent = MockedAgentFactory()

    class DummyBehav(OneShotBehaviour):
        async def run(self):
            self.agent.recv_msg = await self.receive(10)

    template = Template(to="fake@jid")
    behav = DummyBehav()
    agent.add_behaviour(behav, template=template)

    await agent.start(auto_register=False)

    msg_to_send = Message(to="fake@jid", body="BODY", metadata={"performative": "TEST"})
    agent.submit(behav.send(msg_to_send))
    await behav.join()

    assert str(agent.recv_msg.to) == "fake@jid"
    assert agent.recv_msg.body == "BODY"
    assert agent.recv_msg.metadata == {"performative": "TEST"}


async def test_stop_agent_with_blocking_await():
    agent1 = MockedAgentFactory()
    agent1.value = 1000

    class StopBehav(OneShotBehaviour):
        async def run(self):
            await asyncio.sleep(0.5)
            await self.agent.stop()

    class DummyBehav(OneShotBehaviour):
        async def run(self):
            await self.receive(timeout=1000000)
            self.agent.value = 2000

    stopbehah = StopBehav()
    dummybehav = DummyBehav()

    agent1.add_behaviour(dummybehav)
    agent1.add_behaviour(stopbehah)

    await agent1.start(auto_register=False)

    await stopbehah.join()

    assert not agent1.is_alive()
    assert agent1.value == 1000



import asyncio
from unittest.mock import AsyncMock, MagicMock, Mock

from slixmpp import JID

from spade.behaviour import OneShotBehaviour, CyclicBehaviour
from spade.container import Container
from spade.message import Message
from .factories import MockedAgentFactory


async def test_use_container():
    container = Container()
    container.reset()

    agent = MockedAgentFactory()

    assert agent.container == Container()

    assert container.has_agent(str(agent.jid))
    assert container.get_agent(str(agent.jid)) == agent

    await agent.stop()


async def test_send_message_with_container():
    class FakeReceiverAgent:
        def __init__(self):
            self.jid = "fake_receiver_agent@server"

        def set_container(self, c):
            pass

        def set_loop(self, loop):
            pass

        def stop(self):
            pass

        def is_alive(self):
            return False

    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            message = Message(to="fake_receiver_agent@server")
            await self.send(message)
            self.kill()

    container = Container()
    container.reset()
    fake_receiver_agent = FakeReceiverAgent()
    container.register(fake_receiver_agent)

    fake_receiver_agent.dispatch = MagicMock()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    agent.client.send = Mock()
    behaviour = SendBehaviour()
    agent.add_behaviour(behaviour)

    await behaviour.join()

    assert agent.client.send.call_count == 0

    assert fake_receiver_agent.dispatch.call_count == 1
    assert (
        str(fake_receiver_agent.dispatch.call_args[0][0].to)
        == "fake_receiver_agent@server"
    )

    await agent.stop()


async def test_send_message_to_outer_with_container():
    class SendBehaviour(OneShotBehaviour):
        async def run(self):
            message = Message(to="to@outerhost")
            await self.send(message)
            self.kill()

    container = Container()
    container.reset()

    agent = MockedAgentFactory()
    await agent.start(auto_register=False)

    behaviour = SendBehaviour()
    behaviour._xmpp_send = AsyncMock()
    agent.add_behaviour(behaviour)

    await behaviour.join()

    assert container.has_agent(str(agent.jid))
    assert not container.has_agent("to@outerhost")

    assert behaviour._xmpp_send.await_count == 1

    args, kwargs = behaviour._xmpp_send.await_args
    msg_arg = kwargs["msg"]
    assert msg_arg.to == JID("to@outerhost")

    await agent.stop()


def test_unregister():
    container = Container()

    agent = MockedAgentFactory()
    agent2 = MockedAgentFactory(jid="agent2@server")

    assert container.has_agent(str(agent.jid))
    assert container.has_agent(str(agent2.jid))

    container.unregister(agent.jid)

    assert not container.has_agent(str(agent.jid))
    assert container.has_agent(str(agent2.jid))


async def test_cancel_tasks():
    agent = MockedAgentFactory()

    class Behav(CyclicBehaviour):
        async def run(self):
            await asyncio.sleep(100)
            self.has_finished = True

    behav = Behav()
    behav.has_finished = False
    agent.add_behaviour(behaviour=behav)
    await agent.start()

    assert not behav.has_finished

    await agent.stop()

    assert not behav.has_finished



import asyncio
import logging
import sys
import platform
from contextlib import suppress
from typing import Coroutine, Awaitable

from singletonify import singleton

from .behaviour import BehaviourType
from .message import Message

logger = logging.getLogger("SPADE")

# check if python is 3.6 or higher
if sys.version_info >= (3, 7) and sys.platform == "win32":
    asyncio.set_event_loop_policy(
        asyncio.WindowsSelectorEventLoopPolicy()
    )  # pragma: no cover


def get_or_create_eventloop():  # pragma: no cover
    """
    Create or retrieve the appropriate event loop, using uvloop on Linux/macOS
    and winloop on Windows if available.
    """
    # Use uvloop or winloop if available
    if platform.system() == "Windows":
        try:
            import winloop

            asyncio.set_event_loop_policy(winloop.EventLoopPolicy())
        except ImportError:
            pass  # winloop is not available, use default
    else:
        try:
            import uvloop

            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except ImportError:
            pass  # uvloop is not available, use default

    # Create or retrieve the event loop
    if sys.version_info < (3, 10):
        loop = asyncio.get_event_loop()
    else:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
    return loop


@singleton()
class Container(object):
    """
    The container class allows agents to exchange messages
    without using the XMPP socket if they are in the same
    process.
    The container is a singleton.
    """

    def __init__(self):
        self.__agents = {}
        self.loop = get_or_create_eventloop()
        self.is_running = True

    def reset(self) -> None:
        """Empty the container by unregistering all the agents."""
        self.__init__()

    def register(self, agent) -> None:
        """
        Register a new agent.

        Args:
            agent (spade.agent.Agent): the agent to be registered
        """
        self.__agents[str(agent.jid)] = agent
        agent.set_container(self)
        agent.set_loop(self.loop)

    def unregister(self, jid: str) -> None:
        if str(jid) in self.__agents:
            del self.__agents[str(jid)]

    def has_agent(self, jid: str) -> bool:
        """
        Check if an agent is registered in the container.
        Args:
            jid (str): the jid of the agent to be checked.

        Returns:
            bool: wether the agent is or is not registered.
        """
        return jid in self.__agents

    def get_agent(self, jid: str):
        """
        Returns a registered agent
        Args:
            jid (str): the identifier of the agent

        Returns:
            spade.agent.Agent: the agent you were looking for

        Raises:
            KeyError: if the agent is not found
        """
        return self.__agents[jid]

    async def send(self, msg: Message, behaviour: BehaviourType) -> None:
        """
        This method sends the message using the container mechanism
        when the receiver is also registered in the container. Otherwise,
        it uses the XMPP send method from the original behaviour.
        Args:
            msg (spade.message.Message): the message to be sent
            behaviour: the behaviour that is sending the message
        """
        to = str(msg.to)
        if to in self.__agents:
            self.__agents[to].dispatch(msg)
        else:
            await behaviour._xmpp_send(msg=msg)

    def run(self, coro: Awaitable) -> None:  # pragma: no cover
        self.loop.run_until_complete(coro)


def run_container(main_func: Coroutine) -> None:  # pragma: no cover
    container = Container()
    try:
        container.run(main_func)
    except KeyboardInterrupt:
        logger.warning("Keyboard interrupt received. Stopping SPADE...")
    except Exception as e:  # pragma: no cover
        logger.error("Exception in the event loop: {}".format(e))

    # Cancel all pending tasks
    if sys.version_info >= (3, 7):  # pragma: no cover
        tasks = asyncio.all_tasks(loop=container.loop)
    else:
        tasks = asyncio.Task.all_tasks(loop=container.loop)

    for task in tasks:
        task.cancel()
        with suppress(asyncio.CancelledError):
            container.run(task)

    # Shutdown asynchronous generators
    container.loop.run_until_complete(container.loop.shutdown_asyncgens())

    # Close the event loop
    container.loop.close()
    logger.debug("Loop closed")



# coding=utf-8
import datetime
import itertools
from typing import Optional, List

from slixmpp import JID

from .message import Message


def _agent_in_msg(agent, msg):
    to = msg.to.bare if msg.to else None
    sender = msg.sender.bare if msg.sender else None
    return to == agent or sender == agent


class TraceStore(object):
    """Stores and allows queries about events."""

    def __init__(self, size: int):
        self.size = size
        self.store = []

    def reset(self) -> None:
        """Resets the trace store"""
        self.store = []

    def append(self, event: Message, category: Optional[str] = None) -> None:
        """
        Adds a new event to the trace store.
        The event may hava a category

        Args:
          event (spade.message.Message): the event to be stored
          category (str, optional): a category to classify the event (Default value = None)

        """
        date = datetime.datetime.now()
        self.store.insert(0, (date, event, category))
        if len(self.store) > self.size:
            del self.store[-1]

    def len(self) -> int:
        """
        Length of the store

        Returns:
          int: the size of the trace store

        """
        return len(self.store)

    def all(self, limit: Optional[int] = None) -> List[Message]:
        """
        Returns all the events, until a limit if defined

        Args:
          limit (int, optional): the max length of the events to return (Default value = None)

        Returns:
          list: a list of events

        """
        return self.store[:limit][::-1]

    def received(self, limit: Optional[int] = None) -> List[Message]:
        """
        Returns all the events that have been received (excluding sent events), until a limit if defined

        Args:
          limit (int, optional): the max length of the events to return (Default value = None)

        Returns:
          list: a list of received events

        """
        return list(
            itertools.islice(
                (itertools.filterfalse(lambda x: x[1].sent, self.store)), limit
            )
        )[::-1]

    def filter(
        self,
        limit: Optional[int] = None,
        to: Optional[str] = None,
        category: Optional[str] = None,
    ) -> List[Message]:
        """
        Returns the events that match the filters

        Args:
          limit (int, optional): the max length of the events to return (Default value = None)
          to (str, optional): only events that have been sent or received by 'to' (Default value = None)
          category (str, optional): only events belonging to the category (Default value = None)

        Returns:
          list: a list of filtered events

        """
        if category and not to:
            msg_slice = itertools.islice(
                (x for x in self.store if x[2] == category), limit
            )
        elif to and not category:
            to = JID(to)
            msg_slice = itertools.islice(
                (x for x in self.store if _agent_in_msg(to, x[1])), limit
            )
        elif to and category:
            to = JID(to)
            msg_slice = itertools.islice(
                (x for x in self.store if _agent_in_msg(to, x[1]) and x[2] == category),
                limit,
            )
        else:
            msg_slice = self.all(limit=limit)
            return msg_slice

        return list(msg_slice)[::-1]



import pytest
from slixmpp import Presence, JID, Iq
from spade.presence import Contact

from .factories import MockedPresenceAgentFactory
from spade.presence import PresenceManager, PresenceInfo, PresenceShow, PresenceType, ContactNotFound


async def test_presence_info_equals():
    presence1 = PresenceInfo(PresenceType.AVAILABLE, PresenceShow.CHAT, "Online", 5)
    presence2 = PresenceInfo(PresenceType.AVAILABLE, PresenceShow.CHAT, "Online", 5)

    assert presence1 == presence2

async def test_presence_info_not_equals():
    presence1 = PresenceInfo(PresenceType.AVAILABLE, PresenceShow.CHAT, "Online", 5)
    presence2 = PresenceInfo(PresenceType.AVAILABLE, PresenceShow.AWAY, "Online", 5)

    assert presence1 != presence2

async def test_get_state_not_available():
    agent = MockedPresenceAgentFactory(available=False, show=PresenceShow.NONE)
    await agent.start(auto_register=False)

    await agent.start(auto_register=False)

    assert agent.presence.current_presence is None


async def test_presence_manager_initialization(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=False)
    assert len(manager.contacts) == 0


async def test_handle_presence_available(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=False)

    presence = Presence()
    presence['from'] = jid
    presence['type'] = 'available'
    presence['show'] = 'chat'
    presence['status'] = 'Online'
    presence['priority'] = '5'

    manager.handle_presence(presence)

    contact = manager.get_contact(jid.bare)
    assert contact.is_available()
    assert contact.current_presence.show == PresenceShow.CHAT
    assert contact.current_presence.status == 'Online'


async def test_handle_presence_unavailable(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=False)

    # First, set the contact as available
    presence_available = Presence()
    presence_available['from'] = jid
    presence_available['type'] = 'available'
    presence_available['show'] = 'chat'
    presence_available['status'] = 'Online'
    presence_available['priority'] = '5'

    manager.handle_presence(presence_available)

    # Now, set the contact as unavailable
    presence_unavailable = Presence()
    presence_unavailable['from'] = jid
    presence_unavailable['type'] = 'unavailable'

    manager.handle_presence(presence_unavailable)

    contact = manager.get_contact(jid.bare)
    assert not contact.is_available()


async def test_handle_subscription_approve_all(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=True)

    presence_subscribe = Presence()
    presence_subscribe['from'] = jid
    presence_subscribe['type'] = 'subscribe'
    presence_subscribe['ask'] = 'subscribe'
    presence_subscribe['subscription'] = 'subscribe'

    manager.handle_subscription(presence_subscribe)

    contact = manager.get_contact(jid.bare)
    assert contact.subscription == 'subscribed'


async def test_handle_subscription_manual(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=False)

    presence_subscribe = Presence()
    presence_subscribe['from'] = jid
    presence_subscribe['type'] = 'subscribe'
    presence_subscribe['ask'] = 'subscribe'

    manager.handle_subscription(presence_subscribe)

    contact = manager.get_contact(jid.bare)
    assert contact.subscription == 'subscribe'



async def test_set_presence():
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent)
    manager.set_presence(PresenceType.AVAILABLE, PresenceShow.CHAT, "Busy", 10)

    assert manager.current_presence.type == PresenceType.AVAILABLE
    assert manager.current_presence.show == PresenceShow.CHAT
    assert manager.current_presence.status == "Busy"
    assert manager.current_presence.priority == 10



async def test_get_contact_presence(jid):
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent, approve_all=False)

    presence = Presence()
    presence['from'] = jid
    presence['type'] = 'available'
    presence['show'] = 'chat'
    presence['status'] = 'Online'
    presence['priority'] = '5'

    manager.handle_presence(presence)

    presence_info = manager.get_contact_presence(jid.bare)
    assert presence_info.show == PresenceShow.CHAT
    assert presence_info.status == 'Online'
    assert presence_info.priority == 5



async def test_contact_not_found():
    mocked_agent = MockedPresenceAgentFactory()
    await mocked_agent.start()
    manager = PresenceManager(agent=mocked_agent)

    with pytest.raises(ContactNotFound):
        manager.get_contact("nonexistent@localhost")


async def test_handle_roster_update(jid: JID, iq: Iq):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.presence.handle_roster_update(iq)

    contacts = agent.presence.get_contacts()

    bare_jid = jid.bare
    assert bare_jid in contacts
    assert type(contacts[bare_jid]) == Contact
    assert contacts[bare_jid].name == "My Friend"
    assert contacts[bare_jid].subscription == "both"
    assert contacts[bare_jid].groups == ["Friends"]
    assert contacts[bare_jid].ask == "none"
    assert hasattr(contacts[bare_jid], "resources")
    assert isinstance(contacts[bare_jid].resources, dict)


async def test_update_roster(jid: JID):
    agent = MockedPresenceAgentFactory()

    await agent.start(auto_register=False)

    agent.client.update_roster(jid=JID("johndoe@localhost"), name="My Friend")
    
    jid2 = JID("johndoe@localhost")

    iq = agent.client.make_iq_result(ito=jid)
    #set namespace to roster
    iq['roster']['xmlns'] = 'jabber:iq:roster'
    iq['type'] = 'result'
    iq['roster']['items'] = {
        str(jid2): {
            'name': 'My Friend',
            'subscription': 'to',
            'groups': []
        }
    }

    agent.client.event("roster_update", iq)

    stanza = Presence()
    stanza["from"] = jid2
    stanza.set_show(PresenceShow.CHAT.value)
    stanza["status"] = "Just Chatting"
    stanza.set_priority(2)

    agent.client.event("presence_available", stanza)

    contacts = agent.presence.get_contacts()

    bare_jid = jid2.bare
    assert bare_jid in contacts
    assert type(contacts[bare_jid]) == Contact
    assert contacts[bare_jid].name == "My Friend"
    assert contacts[bare_jid].subscription == "to"
    assert contacts[bare_jid].groups == []
    assert contacts[bare_jid].ask == "none"
    assert hasattr(contacts[bare_jid], "resources")
    assert isinstance(contacts[bare_jid].resources, dict)
    assert contacts[bare_jid].resources[jid2.resource].type == PresenceType.AVAILABLE
    assert contacts[bare_jid].resources[jid2.resource].show == PresenceShow.CHAT
    assert contacts[bare_jid].resources[jid2.resource].status == "Just Chatting"
    assert contacts[bare_jid].resources[jid2.resource].priority == 2


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `message` package."""
import copy

from slixmpp.plugins.xep_0004 import Form
from slixmpp import JID
from slixmpp import Message as slixmppMessage
import pytest

from spade.message import Message, SPADE_X_METADATA


def test_prepare(message):
    aiomsg = message.prepare()
    assert aiomsg['to'] == JID("to@localhost")
    assert aiomsg['from'] == JID("sender@localhost")
    assert aiomsg['body'] == "message body"

    for form in [pl for pl in aiomsg.get_payload() if pl.tag == '{jabber:x:data}x']:
        if form.find('{jabber:x:data}title').text == SPADE_X_METADATA:
            for field in form.findall('{jabber:x:data}field'):
                if field.attrib['var'] == "_thread_node":
                    assert field.find('{jabber:x:data}value').text == "thread-id"
                else:
                    assert message.get_metadata(field.attrib['var']) == field.find('{jabber:x:data}value').text


def test_message_to_string(message):
    assert message.__str__() == ('<message to="to@localhost" from="sender@localhost" thread="thread-id" metadata={'
                                 '\'metadata1\': \'value1\', \'metadata2\': \'value2\'}>\nmessage body\n</message>')


def test_make_reply(message):
    reply = message.make_reply()
    assert reply.to == JID("sender@localhost")
    assert reply.sender == JID("to@localhost")
    assert reply.body == "message body"
    assert reply.thread == "thread-id"
    assert reply.get_metadata("metadata1") == "value1"
    assert reply.get_metadata("metadata2") == "value2"


def test_message_from_node_attribute_error():
    with pytest.raises(AttributeError):
        Message.from_node(Message())


def test_body_with_languages():
    msg = slixmppMessage()
    msg.chat()
    msg['body'] = {
        'en': 'Hello World',
        'es': 'Hola Mundo'
    }

    new_msg = Message.from_node(msg)
    assert new_msg.body == "Hello World"

    msg = slixmppMessage()
    msg.chat()


def test_message_from_node():
    slimsg = slixmppMessage()
    slimsg.chat()

    data = Form()
    data['type'] = 'form'

    data.add_field(
        var="performative",
        ftype='text-single',
        value="request"
    )

    data.add_field(
        var="_thread_node",
        ftype='text-single',
        value="thread-id"
    )

    data['title'] = SPADE_X_METADATA
    slimsg.append(data)

    msg = Message.from_node(slimsg)

    assert msg.thread == "thread-id"
    assert msg.get_metadata("performative") == "request"
    assert msg.metadata == {"performative": "request"}


def test_thread_empty():
    msg = Message(thread=None)

    assert msg.thread is None
    assert msg.metadata == {}

    slimsg = msg.prepare()
    for data in [pl for pl in slimsg.get_payload() if pl.tag == '{jabber:x:data}x']:
        if data.findall('{jabber:x:data}title').text == SPADE_X_METADATA:
            for field in data.findall('{jabber:x:data}field'):
                assert field.attrib['var'] != "_thread_node"


def test_equal(message):
    assert message == copy.copy(message)


def test_not_equal(message, message2):
    assert message != message2


def test_id(message):
    assert isinstance(message.id, int)


def test_metadata_is_string():
    Message(metadata={"key": "value"})


def test_metadata_is_not_string():
    with pytest.raises(TypeError):
        Message(metadata={"key": 1000})


def test_metadata_set_string():
    msg = Message()
    msg.set_metadata("key", "value")


def test_metadata_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.set_metadata(1000, "value")


def test_body_is_string():
    Message(body="body")


def test_body_is_not_string():
    with pytest.raises(TypeError):
        Message(body={})


def test_body_set_string():
    msg = Message()
    msg.body = "body"


def test_body_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.body = 1000


def test_body_set_none():
    msg = Message()
    msg.body = None


def test_to_is_string():
    Message(to="agent@fakeserver")


def test_to_is_not_string():
    with pytest.raises(TypeError):
        Message(to=JID("agent@fakeserver"))


def test_to_set_string():
    msg = Message()
    msg.to = "agent@fakeserver"


def test_to_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.to = 1000


def test_to_set_none():
    msg = Message()
    msg.to = None


def test_sender_is_string():
    Message(sender="agent@fakeserver")


def test_sender_is_not_string():
    with pytest.raises(TypeError):
        Message(sender=JID("agent@fakeserver"))


def test_sender_set_string():
    msg = Message()
    msg.sender = "agent@fakeserver"


def test_sender_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.sender = 1000


def test_sender_set_none():
    msg = Message()
    msg.sender = None


def test_thread_is_string():
    Message(thread="thread_id_001")


def test_thread_is_not_string():
    with pytest.raises(TypeError):
        Message(thread=1000)


def test_thread_set_string():
    msg = Message()
    msg.thread = "thread_id_001"


def test_thread_set_not_string():
    msg = Message()
    with pytest.raises(TypeError):
        msg.thread = 1000


def test_thread_set_none():
    msg = Message()
    msg.thread = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `template` package."""
import pytest

from spade.message import Message
from spade.template import Template


def test_match():
    template = Template()
    template.sender = "sender1@host"
    template.to = "recv1@host"
    template.body = "Hello World"
    template.thread = "thread-id"
    template.metadata = {"performative": "query"}

    message = Message()
    message.sender = "sender1@host"
    message.to = "recv1@host"
    message.body = "Hello World"
    message.thread = "thread-id"
    message.set_metadata("performative", "query")

    assert template.match(message)


def test_match_false_sender():
    template = Template()
    template.sender = "sender2@host"

    message = Message()

    assert not template.match(message)

    message.sender = "sender1@host"

    assert not template.match(message)


def test_match_false_to():
    template = Template()
    template.to = "recv1@host"

    message = Message()

    assert not template.match(message)

    message.to = "recv2@host"

    assert not template.match(message)


def test_match_false_body():
    template = Template()
    template.body = "Hello World"

    message = Message()

    assert not template.match(message)

    message.body = "Bye Bye Love"

    assert not template.match(message)


def test_match_false_thread():
    template = Template()
    template.thread = "thread-id"

    message = Message()

    assert not template.match(message)

    message.thread = "thread-id-false"

    assert not template.match(message)


def test_match_false_metadata():
    template = Template()
    template.metadata = {"performative": "query"}

    message = Message()

    assert not template.match(message)

    message.set_metadata("performative", "inform")

    assert not template.match(message)


def test_match_false_metadata_with_different_key():
    template = Template()
    template.metadata = {"performative": "query"}

    message = Message()
    message.set_metadata("language", "query")

    assert not template.match(message)


def test_match_and():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    assert not (t1 & t2).match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert not (t1 & t2).match(m2)

    m3 = Message()
    m3.sender = "sender1@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "query"}

    assert (t1 & t2).match(m3)


def test_match_iand():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    t1 &= t2
    assert not t1.match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert not t1.match(m2)

    m3 = Message()
    m3.sender = "sender1@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "query"}

    assert t1.match(m3)


def test_match_or():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    assert (t1 | t2).match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert (t1 | t2).match(m2)

    m3 = Message()
    m3.sender = "sender2@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "inform"}

    assert not (t1 | t2).match(m3)


def test_match_ior():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    t1 |= t2

    assert t1.match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert t1.match(m2)

    m3 = Message()
    m3.sender = "sender2@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "inform"}

    assert not t1.match(m3)


def test_match_xor():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    assert (t1 ^ t2).match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert (t1 ^ t2).match(m2)

    m3 = Message()
    m3.sender = "sender2@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "inform"}

    assert not (t1 ^ t2).match(m3)

    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.sender = "sender1@host"
    m4 = Message()
    m4.sender = "sender1@host"

    assert not (t1 ^ t2).match(m4)


def test_match_ixor():
    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.to = "recv1@host"
    t2.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    t1 ^= t2

    assert t1.match(m1)

    m2 = Message()
    m2.to = "recv1@host"
    m2.metadata = {"performative": "query"}

    assert t1.match(m2)

    m3 = Message()
    m3.sender = "sender2@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "inform"}

    assert not t1.match(m3)

    t1 = Template()
    t1.sender = "sender1@host"
    t2 = Template()
    t2.sender = "sender1@host"
    m4 = Message()
    m4.sender = "sender1@host"

    t1 ^= t2

    assert not t1.match(m4)


def test_match_not():
    t1 = Template()
    t1.sender = "sender1@host"
    t1.to = "recv1@host"
    t1.metadata = {"performative": "query"}

    m1 = Message()
    m1.sender = "sender1@host"

    assert (~t1).match(m1)

    m2 = Message()
    m2.sender = "sender1@host"
    m2.to = "recv1@host"

    assert (~t1).match(m2)

    m3 = Message()
    m3.sender = "sender1@host"
    m3.to = "recv1@host"
    m3.metadata = {"performative": "query"}

    assert not (~t1).match(m3)


def test_and_value_error():
    with pytest.raises(TypeError):
        assert Template() & None

    with pytest.raises(TypeError):
        assert Template() & 3

    with pytest.raises(TypeError):
        assert Template() & "string"

    with pytest.raises(TypeError):
        assert Template() & object()


def test_or_value_error():
    with pytest.raises(TypeError):
        assert Template() | None

    with pytest.raises(TypeError):
        assert Template() | 3

    with pytest.raises(TypeError):
        assert Template() | "string"

    with pytest.raises(TypeError):
        assert Template() | object()


def test_xor_value_error():
    with pytest.raises(TypeError):
        assert Template() ^ None

    with pytest.raises(TypeError):
        assert Template() ^ 3

    with pytest.raises(TypeError):
        assert Template() ^ "string"

    with pytest.raises(TypeError):
        assert Template() ^ object()



import logging
from abc import ABCMeta

from .message import MessageBase

logger = logging.getLogger("spade.Template")


# TODO: Include regex in templates


class BaseTemplate(metaclass=ABCMeta):
    """Template operators"""

    def __and__(self, other):
        """Implementation of & operator"""
        if not issubclass(other.__class__, BaseTemplate):
            raise TypeError("Expressions must be of class Template")
        return ANDTemplate(self, other)

    def __iand__(self, other):
        """Implementation of &= operator"""
        return self & other

    def __or__(self, other):
        """Implementation of | operator"""
        if not issubclass(other.__class__, BaseTemplate):
            raise TypeError("Expressions must be of class Template")
        return ORTemplate(self, other)

    def __ior__(self, other):
        """Implementation of |= operator"""
        return self | other

    def __xor__(self, other):
        """Implementation of ^ operator"""
        if not issubclass(other.__class__, BaseTemplate):
            raise TypeError("Expressions must be of class Template")
        return XORTemplate(self, other)

    def __ixor__(self, other):
        """Implementation of ^= operator"""
        return self ^ other

    def __invert__(self):
        """Implementation of ~ operator"""
        return NOTTemplate(self)


class NOTTemplate(BaseTemplate):
    """ """

    def __init__(self, expr):
        self.expr = expr

    def match(self, message):
        """ """
        return not (self.expr.match(message))


class ORTemplate(BaseTemplate):
    """ """

    def __init__(self, expr1, expr2):
        self.expr1 = expr1
        self.expr2 = expr2

    def match(self, message):
        """ """
        return self.expr1.match(message) | self.expr2.match(message)


class ANDTemplate(BaseTemplate):
    """ """

    def __init__(self, expr1, expr2):
        self.expr1 = expr1
        self.expr2 = expr2

    def match(self, message):
        """ """
        return self.expr1.match(message) & self.expr2.match(message)


class XORTemplate(BaseTemplate):
    """ """

    def __init__(self, expr1, expr2):
        self.expr1 = expr1
        self.expr2 = expr2

    def match(self, message):
        """ """
        return self.expr1.match(message) ^ self.expr2.match(message)


class Template(BaseTemplate, MessageBase):
    """Template for message matching"""

    def __str__(self) -> str:
        s = f'<template to="{self.to}" from="{self.sender}" thread="{self.thread}" metadata={self.metadata}>'
        if self.body:
            s += "\n" + self.body + "\n"
        s += "</template>"
        return s



import datetime
import itertools
from collections import namedtuple

from slixmpp import JID

from spade.message import Message
from spade.trace import TraceStore, _agent_in_msg


def test_agent_in_msg_sender():
    msg = Message(sender="agent@server")
    assert _agent_in_msg(JID("agent@server"), msg)


def test_agent_in_msg_to():
    msg = Message(to="agent@server")
    assert _agent_in_msg(JID("agent@server"), msg)


def test_agent_in_msg_false():
    msg = Message(to="agent2@server")
    assert not _agent_in_msg(JID("agent@server"), msg)


def test_init():
    trace = TraceStore(10)

    assert trace.size == 10
    assert trace.store == []


def test_reset():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)
    assert len(trace.store) == 5

    trace.reset()
    assert len(trace.store) == 0


def test_append():
    trace = TraceStore(10)
    trace.append("EVENT")
    assert len(trace.store) == 1
    assert trace.store[0][1] == "EVENT"
    assert type(trace.store[0][0]) == datetime.datetime
    assert trace.store[0][2] is None


def test_append_to_top():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)
    trace.append("EVENT")
    assert len(trace.store) == 6
    assert trace.store[0][1] == "EVENT"


def test_append_with_category():
    trace = TraceStore(10)
    trace.append("EVENT", "CATEGORY")
    assert len(trace.store) == 1
    assert trace.store[0][1] == "EVENT"
    assert type(trace.store[0][0]) == datetime.datetime
    assert trace.store[0][2] == "CATEGORY"


def test_append_max_size():
    trace = TraceStore(2)
    for i in range(5):
        trace.append(i)

    assert len(trace.store) == 2
    assert trace.store[0][1] == 4
    assert trace.store[1][1] == 3


def test_len():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)
    assert trace.len() == 5


def test_all():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)
    all = trace.all()
    events = [e[1] for e in all]

    assert events == [0, 1, 2, 3, 4]


def test_all_limit():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)
    all = trace.all(limit=2)
    events = [e[1] for e in all]

    assert events == [3, 4]


def test_received():
    Event = namedtuple("Event", ["value", "sent"])
    trace = TraceStore(10)
    for i in range(5):
        trace.append(Event(i, True))

    empty_received = trace.received()
    assert len(empty_received) == 0

    for i in range(5, 10):
        trace.append(Event(i, False))

    received = [r[1].value for r in trace.received()]
    assert len(received) == 5
    assert received == [5, 6, 7, 8, 9]

    limit_received = [r[1].value for r in trace.received(limit=3)]
    assert len(limit_received) == 3
    assert limit_received == [7, 8, 9]


def test_filter():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)

    result = [x[1] for x in trace.filter()]
    assert result == [0, 1, 2, 3, 4]


def test_filter_limit():
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i)

    result = [x[1] for x in trace.filter(limit=3)]
    assert result == [2, 3, 4]


def test_filter_to():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        msg = Message(
            body=str(i), to="{}@server".format(next(cycle)), sender="sender@server"
        )
        trace.append(msg)

    result = [x[1].body for x in trace.filter(to="0@server")]
    assert result == ["0", "3"]


def test_filter_to_limit():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        msg = Message(
            body=str(i), to="{}@server".format(next(cycle)), sender="sender@server"
        )
        trace.append(msg)

    result = [x[1].body for x in trace.filter(to="0@server", limit=1)]
    assert result == ["3"]


def test_filter_category():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        trace.append(i, str(next(cycle)))

    result = [x[2] for x in trace.filter(category="0")]
    assert result == ["0", "0"]


def test_filter_category_limit():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        trace.append(1, str(next(cycle)))

    result = [x[2] for x in trace.filter(category="0", limit=1)]
    assert result == ["0"]


def test_filter_to_and_category():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        c = str(next(cycle))
        msg = Message(body=str(c), to="{}@server".format(c), sender="sender@server")
        trace.append(msg, c)

    result = [(x[1].body, x[2]) for x in trace.filter(to="1@server", category="1")]
    assert result == [("1", "1"), ("1", "1")]


def test_filter_to_and_category_limit():
    cycle = itertools.cycle(range(3))
    trace = TraceStore(10)
    for i in range(5):
        c = str(next(cycle))
        msg = Message(body=str(c), to="{}@server".format(c), sender="sender@server")
        trace.append(msg, c)

    result = [
        (x[1].body, x[2]) for x in trace.filter(to="1@server", category="1", limit=1)
    ]
    assert result == [("1", "1")]

